import unittest

from django.test.utils import override_settings
from django.urls import resolve
from test_plus import TestCase

from tournamentcontrol.competition.models import Match
from tournamentcontrol.competition.tests import factories


@override_settings(ROOT_URLCONF="tournamentcontrol.competition.tests.urls")
class APITests(TestCase):

    @classmethod
    def setUpTestData(cls):
        """Create test data once for all tests."""
        # Create a club
        cls.club = factories.ClubFactory.create()

        # Create a competition with a known slug
        cls.competition = factories.CompetitionFactory.create()

        # Create a season with a known slug
        cls.season = factories.SeasonFactory.create(competition=cls.competition)

        # Create a division with a known slug
        cls.division = factories.DivisionFactory.create(season=cls.season)

        # Create a stage
        cls.stage = factories.StageFactory.create(division=cls.division)

        # Create teams
        cls.team1 = factories.TeamFactory.create(club=cls.club, division=cls.division)
        cls.team2 = factories.TeamFactory.create(division=cls.division)

        # Create a person for players endpoints
        cls.person = factories.PersonFactory.create(club=cls.club)
        cls.team_association = factories.TeamAssociationFactory.create(
            team=cls.team1, person=cls.person
        )

        # Create a venue and ground for matches
        cls.venue = factories.VenueFactory.create(season=cls.season)
        cls.ground = factories.GroundFactory.create(venue=cls.venue)

        # Create a match
        cls.match = factories.MatchFactory.create(
            stage=cls.stage,
            home_team=cls.team1,
            away_team=cls.team2,
            play_at=cls.ground,
        )

    def test_club_list(self):
        """Test club-list endpoint"""
        self.get("v1:club-list")
        self.response_200()

    def test_club_detail(self):
        """Test club-detail endpoint"""
        self.get("v1:club-detail", slug=self.club.slug)
        self.response_200()

    def test_competition_list(self):
        """Test competition-list endpoint"""
        self.get("v1:competition-list")
        self.response_200()

    def test_competition_detail(self):
        """Test competition-detail endpoint"""
        self.get("v1:competition-detail", slug=self.competition.slug)
        self.response_200()

    @unittest.skip(
        'Could not resolve URL for hyperlinked relationship using view name "season-detail"'
    )
    def test_season_list(self):
        """Test season-list endpoint"""
        self.get(
            "v1:season-list",
            competition_slug=self.competition.slug,
        )
        self.response_200()

    @unittest.skip(
        'Could not resolve URL for hyperlinked relationship using view name "season-detail"'
    )
    def test_season_detail(self):
        """Test season-detail endpoint"""
        self.get(
            "v1:season-detail",
            competition_slug=self.competition.slug,
            slug=self.season.slug,
        )
        self.response_200()

    def test_division_list(self):
        """Test division-list endpoint"""
        self.get(
            "v1:division-list",
            competition_slug=self.competition.slug,
            season_slug=self.season.slug,
        )
        self.response_200()

    def test_division_detail(self):
        """Test division-detail endpoint"""
        self.get(
            "v1:division-detail",
            competition_slug=self.competition.slug,
            season_slug=self.season.slug,
            slug=self.division.slug,
        )
        self.response_200()

        # Validate that the response contains the expected data, including timezone serialization.
        self.assertJSONEqual(
            self.last_response.content,
            {
                "title": self.division.title,
                "slug": self.division.slug,
                "url": f"http://testserver/api/v1/competitions/{self.competition.slug}/seasons/{self.season.slug}/divisions/{self.division.slug}/",
                "teams": [
                    {
                        "id": self.team1.id,
                        "title": self.team1.title,
                        "slug": self.team1.slug,
                        "club": {
                            "abbreviation": self.club.abbreviation,
                            "facebook": self.club.facebook,
                            "short_title": self.club.short_title,
                            "slug": self.club.slug,
                            "title": self.club.title,
                            "twitter": self.club.twitter,
                            "url": f"http://testserver/api/v1/clubs/{self.club.slug}/",
                            "website": self.club.website,
                            "youtube": self.club.youtube,
                        },
                    },
                    {
                        "id": self.team2.id,
                        "title": self.team2.title,
                        "slug": self.team2.slug,
                        "club": {
                            "abbreviation": self.team2.club.abbreviation,
                            "facebook": self.team2.club.facebook,
                            "short_title": self.team2.club.short_title,
                            "slug": self.team2.club.slug,
                            "title": self.team2.club.title,
                            "twitter": self.team2.club.twitter,
                            "url": f"http://testserver/api/v1/clubs/{self.team2.club.slug}/",
                            "website": self.team2.club.website,
                            "youtube": self.team2.club.youtube,
                        },
                    },
                ],
                "stages": [
                    {
                        "title": self.stage.title,
                        "slug": self.stage.slug,
                        "url": f"http://testserver/api/v1/competitions/{self.competition.slug}/seasons/{self.season.slug}/divisions/{self.division.slug}/stages/{self.stage.slug}/",
                        "matches": [
                            {
                                "id": self.match.id,
                                "uuid": str(self.match.uuid),
                                "round": f"Round {self.match.round}",
                                "date": self.match.date.isoformat(),
                                "time": self.match.time.isoformat(),
                                "datetime": self.match.datetime.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
                                "is_bye": False,
                                "is_washout": False,
                                "home_team": self.team1.id,
                                "home_team_score": None,
                                "away_team": self.team2.id,
                                "away_team_score": None,
                                "referees": [],
                                "videos": None,
                                "play_at": {
                                    "id": self.ground.id,
                                    "title": self.ground.title,
                                    "abbreviation": None,
                                    "timezone": str(self.ground.timezone),
                                },
                            },
                        ],
                    }
                ],
            },
        )

    def test_players_list(self):
        """Test players-list endpoint"""
        self.get(
            "v1:players-list",
            competition_slug=self.competition.slug,
            season_slug=self.season.slug,
        )
        self.response_200()

    def test_players_detail(self):
        """Test players-detail endpoint"""
        self.get(
            "v1:players-detail",
            competition_slug=self.competition.slug,
            season_slug=self.season.slug,
            id=self.team_association.pk,
        )
        self.response_200()

    def test_match_list(self):
        """Test match-list endpoint"""
        self.get(
            "v1:match-list",
            competition_slug=self.competition.slug,
            season_slug=self.season.slug,
        )
        self.response_200()

    def test_match_detail(self):
        """Test match-detail endpoint"""
        self.get(
            "v1:match-detail",
            competition_slug=self.competition.slug,
            season_slug=self.season.slug,
            uuid=self.match.uuid,
        )
        self.response_200()

    def test_stage_list(self):
        """Test stage-list endpoint"""
        self.get(
            "v1:stage-list",
            competition_slug=self.competition.slug,
            season_slug=self.season.slug,
            division_slug=self.division.slug,
        )
        self.response_200()

    def test_stage_detail(self):
        """Test stage-detail endpoint"""
        self.get(
            "v1:stage-detail",
            competition_slug=self.competition.slug,
            season_slug=self.season.slug,
            division_slug=self.division.slug,
            slug=self.stage.slug,
        )
        self.response_200()



import unittest

from django.test.utils import override_settings
from example_app.models import TestDateTimeField
from example_app.tests.factories import TestDateTimeFieldFactory
from test_plus import TestCase as BTC

from touchtechnology.common.tests import factories


@override_settings(ROOT_URLCONF="example_app.urls")
class TestCase(BTC):
    def setUp(self):
        self.staff = factories.UserFactory.create(is_staff=True, is_superuser=True)
        self.regular = factories.UserFactory.create()


class GenericListTests(TestCase):
    def test_vanilla(self):
        self.assertGoodView("generic:vanilla:list")
        self.assertResponseContains("<title>test date time field</title>")
        self.assertResponseContains("<h1>test date time fields</h1>")

    def test_permissions_403(self):
        "Ensure unprivileged user is denied, superuser may view"
        self.assertLoginRequired("generic:permissions:list")

        with self.login(self.regular):
            self.get("generic:permissions:list")
            self.response_403()

        with self.login(self.staff):
            self.assertGoodView("generic:permissions:list")


class GenericDetailTests(TestCase):
    def setUp(self):
        super().setUp()
        TestDateTimeFieldFactory.reset_sequence()
        TestDateTimeFieldFactory.create_batch(20)
        self.instance = TestDateTimeField.objects.earliest("pk")

    def test_vanilla(self):
        self.assertGoodView("generic:vanilla:detail", self.instance.pk)
        self.assertResponseContains("<title>test date time field</title>")
        self.assertResponseContains("<h1>test date time fields</h1>")
        self.assertResponseContains(f"<p>{self.instance.datetime}</p>")

    @unittest.skip("permissions_required not implemented for generic_detail")
    def test_permissions_403(self):
        self.get("generic:permissions:detail", self.instance.pk)
        self.response_403()


class GenericEditTests(TestCase):
    def test_vanilla(self):
        self.assertGoodView("generic:vanilla:add")
        self.assertResponseContains("<title>Edit test date time field</title>")

        data = {
            "datetime_0": "2015-1-1",
            "datetime_1": "9:00",
            "datetime_2": "UTC",
        }
        res = self.post("generic:vanilla:add", data=data)
        self.assertEqual(1, TestDateTimeField.objects.count())
        self.response_302()
        self.assertRedirects(res, self.reverse("generic:vanilla:list"))

        self.assertGoodView("generic:vanilla:list")
        self.assertResponseContains("<li>2015-01-01 09:00:00+00:00</li>")

    def test_permissions_403(self):
        "Ensure unprivileged user is denied, superuser may view and edit"
        self.assertLoginRequired("generic:permissions:add")

        data = {
            "datetime_0": "2015-1-1",
            "datetime_1": "9:00",
            "datetime_2": "UTC",
        }

        with self.login(self.regular):
            self.get("generic:permissions:add")
            self.response_403()
            self.post("generic:permissions:add", data=data)
            self.response_403()

        with self.login(self.staff):
            self.assertGoodView("generic:permissions:add")
            self.post("generic:permissions:add", data=data)
            self.response_302()


class GenericDeleteTests(TestCase):
    def setUp(self):
        super(GenericDeleteTests, self).setUp()
        self.object = TestDateTimeFieldFactory.create()

    def test_vanilla(self):
        self.post("generic:vanilla:delete", self.object.pk)
        self.response_302()

    def test_permissions_403(self):
        "Ensure unprivileged user is denied, superuser may delete"
        self.post("generic:permissions:delete", self.object.pk)
        self.response_302()  # Anonymous users will be redirected to login

        with self.login(self.regular):
            self.post("generic:permissions:delete", self.object.pk)
            self.response_403()

        with self.login(self.staff):
            self.post("generic:permissions:delete", self.object.pk)
            self.response_302()


class GenericPermissionsTests(TestCase):
    def setUp(self):
        super(GenericPermissionsTests, self).setUp()
        self.object = TestDateTimeFieldFactory.create()

    def test_vanilla(self):
        self.get("generic:vanilla:perms", self.object.pk)
        self.response_403()

        with self.login(self.regular):
            self.get("generic:vanilla:perms", self.object.pk)
            self.response_403()

        with self.login(self.staff):
            self.assertGoodView("generic:vanilla:perms", self.object.pk)
            self.response_200()
            self.assertResponseContains(
                """
                <p>
                    Control who <em>can change test date time field</em>
                    for the test date time field
                    <strong>{}</strong>.
                </p>
            """.format(
                    self.object
                )
            )


class GenericEditMultipleTests(TestCase):
    def test_vanilla(self):
        self.assertGoodView("generic:vanilla:edit")
        self.assertResponseContains(
            "<title>Edit multiple test date time fields</title>"
        )

        data = {
            # management form hidden fields
            "form-TOTAL_FORMS": 2,
            "form-INITIAL_FORMS": 0,
            "form-MIN_NUM_FORMS": 0,
            "form-MAX_NUM_FORMS": 1000,
            "form-0-datetime_0": "2015-1-1",
            "form-0-datetime_1": "9:00",
            "form-0-datetime_2": "UTC",
            "form-1-datetime_0": "2016-3-3",
            "form-1-datetime_1": "9:00",
            "form-1-datetime_2": "UTC",
        }
        res = self.post("generic:vanilla:edit", data=data)
        self.assertEqual(2, TestDateTimeField.objects.count())
        self.response_302()
        self.assertRedirects(res, self.reverse("generic:vanilla:list"))

        self.assertGoodView("generic:vanilla:list")
        self.assertResponseContains(
            "<ul>"
            "<li>2015-01-01 09:00:00+00:00</li>"
            "<li>2016-03-03 09:00:00+00:00</li>"
            "</ul>"
        )

    def test_permissions_403(self):
        "Ensure unprivileged user is denied, superuser may view and edit"
        self.assertLoginRequired("generic:permissions:edit")

        data = {
            # management form hidden fields
            "form-TOTAL_FORMS": 2,
            "form-INITIAL_FORMS": 0,
            "form-MIN_NUM_FORMS": 0,
            "form-MAX_NUM_FORMS": 1000,
            "form-0-datetime_0": "2015-1-1",
            "form-0-datetime_1": "9:00",
            "form-0-datetime_2": "UTC",
            "form-1-datetime_0": "2016-3-3",
            "form-1-datetime_1": "9:00",
            "form-1-datetime_2": "UTC",
        }

        with self.login(self.regular):
            self.get("generic:permissions:edit")
            self.response_403()
            self.post("generic:permissions:edit", data=data)
            self.response_403()

        with self.login(self.staff):
            self.assertGoodView("generic:permissions:edit")
            self.post("generic:permissions:edit", data=data)
            self.response_302()

        self.assertEqual(2, TestDateTimeField.objects.count())



# -*- coding: utf-8 -*-

from django.db import migrations

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0029_seasonreferee"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="seasonreferee",
            options={
                "ordering": (
                    "club",
                    "season",
                    "person__last_name",
                    "person__first_name",
                ),
                "verbose_name": "referee",
            },
        ),
        migrations.AddField(
            model_name="match",
            name="referees",
            field=touchtechnology.common.db.models.ManyToManyField(
                blank=True, related_name="matches", to="competition.SeasonReferee"
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0003_division_draft_data"),
    ]

    operations = [
        migrations.AlterField(
            model_name="person",
            name="gender",
            field=models.CharField(
                max_length=1,
                choices=[(b"M", "Male"), (b"F", "Female"), (b"X", "Unspecified")],
            ),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("content", "0003_auto_20160831_0608"),
        ("common", "0005_sitemapnode_populate_kwargs"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="placeholderkeywordargument",
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name="placeholderkeywordargument",
            name="node",
        ),
        migrations.DeleteModel(
            name="PlaceholderKeywordArgument",
        ),
    ]



from django.utils.encoding import smart_str
from test_plus import TestCase

from touchtechnology.common.tests import factories


class SitemapNodeTests(TestCase):

    def setUp(self):
        self.object = factories.SitemapNodeFactory.create()

    def test_string_representation(self):
        self.assertEqual(self.object.title, smart_str(self.object))



from django.urls import path
from django.views.i18n import set_language

from touchtechnology.common.views import set_timezone

urlpatterns = [
    path("i18n/set-language/", set_language, name="set-language"),
    path("i18n/set-timezone/", set_timezone, name="set-timezone"),
]



# -*- coding: utf-8 -*-

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0005_alter_club_facebook_youtube_position"),
    ]

    operations = [
        migrations.AddField(
            model_name="division",
            name="sportingpulse_url",
            field=models.URLField(
                help_text="Here be dragons! Enter at own risk!",
                max_length=1024,
                null=True,
                blank=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="match",
            name="external_identifier",
            field=models.CharField(
                db_index=True, max_length=20, unique=True, null=True, blank=True
            ),
            preserve_default=True,
        ),
    ]



import collections
import logging

from django.conf import settings
from django.contrib.admin.sites import (
    AdminSite as DjangoAdminSite,
    AlreadyRegistered,
    NotRegistered,
)
from django.db import connection
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.urls import path, re_path, reverse_lazy
from django.utils.encoding import smart_str
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from touchtechnology.common.decorators import staff_login_required_m
from touchtechnology.common.models import SitemapNode

logger = logging.getLogger(__name__)


class AdminSite(DjangoAdminSite):
    def __init__(self, name="admin"):
        super(AdminSite, self).__init__(name)
        self._components = collections.OrderedDict()
        self._schemas = collections.defaultdict(lambda: None)

    def register(self, component, *schemas):
        logger.debug("Registering admin component... %s" % component.__name__)
        if component in self._components:
            raise AlreadyRegistered(
                _("The component %s is already " "registered" % component.__name__)
            )
        self._components[component] = component(self.name)
        if schemas:
            self._schemas[component] = schemas

    def unregister(self, component):
        self._schemas.pop(component, None)
        try:
            self._components.pop(component)
        except KeyError:
            raise NotRegistered(
                _("The component %s is not " "registered" % component.__name__)
            )

    def _hidden_component(self, component):
        if "tenant_schemas" not in settings.INSTALLED_APPS:
            # Only hide components when we are not in multi-tenant mode
            return False

        get_public_schema_name = import_string(
            "tenant_schemas.utils.get_public_schema_name"
        )

        # For components only registered for specific schemas, return solely
        # based on that option. That is, you can't disable an explicitly
        # registered component for a given schema.
        schemas = self._schemas[component.__class__]
        if schemas:
            if connection.tenant.schema_name in schemas:
                return False
            return True

        application = component.__module__.split(".admin", 1)[0]
        logger.debug("application=%r", application)

        # Before we allow it through, make sure the application is available
        # on this tenant.
        if application in settings.INSTALLED_APPS:
            if connection.tenant.schema_name == get_public_schema_name():
                if application not in settings.SHARED_APPS:
                    return True
            else:
                if application not in settings.TENANT_APPS:
                    return True

        # Lastly, check if the client allows the application in this tenant.
        try:
            return connection.tenant.hidden_component(component)
        except AttributeError:
            return False

    def get_urls(self):
        urlpatterns = [
            path("", self.index, name="index"),
            re_path(
                r"^sitemapnode/move/(?P<id>\d+)/(?P<direction>(up|down))/$",
                self.reorder,
                name="reorder",
            ),
        ]

        for klass, component in self._components.items():
            # It appears that it is not possible to "hide" a component from
            # the admin URL generation due to this being loaded at startup.
            #
            # Instead we could move this into a middleware like the
            # content.SitemapNodeMiddleware but it might require quite lots of
            # refactoring.
            urlpatterns += [
                path(
                    component.name + "/",
                    component.urls,
                    {"admin": self, "component": component},
                ),
            ]

        return urlpatterns

    def get_components(self, show_all=False):
        components = []
        for n, c in self._components.items():
            visible = c.visible or not show_all
            if visible and not self._hidden_component(c):
                components.append(
                    (
                        smart_str(c.verbose_name),
                        c.app_name,
                        smart_str(c.reverse("index")),
                        c,
                        self._schemas[n],
                    )
                )
        return components

    def dashboard(self):
        widgets = ()
        for __, __, __, component, schemas in self.get_components():
            try:
                widgets += component.widgets
            except AttributeError:
                pass
        return widgets

    def get_absolute_url(self):
        return self.reverse("index")

    def reverse(self, name, args=None, kwargs=None, prefix=None):
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        named_url = "%s:%s" % (self.name, name)
        if prefix:
            named_url = "%s:%s" % (prefix, named_url)
        return reverse_lazy(named_url, args=args, kwargs=kwargs, current_app=self.name)

    @staff_login_required_m
    def index(self, request, *args, **kwargs):
        context = {"admin": self}
        context.update(kwargs)
        response = TemplateResponse(
            request, "touchtechnology/admin/index.html", context=context
        )
        response.current_app = self.name
        return response

    @staff_login_required_m
    def reorder(self, request, id, direction, *args, **kwargs):
        node = get_object_or_404(SitemapNode, pk=id)
        getattr(node, "move_%s" % direction)()
        redirect_to = request.META.get("HTTP_REFERER", self.reverse("index"))
        return HttpResponseRedirect(redirect_to)



from test_plus import TestCase

from touchtechnology.common.tests.factories import UserFactory
from tournamentcontrol.competition.models import Match
from tournamentcontrol.competition.tests import factories


class DrawGenerationMatchFormSetTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.superuser = UserFactory.create(is_staff=True, is_superuser=True)

        factories.DrawFormatFactory.reset_sequence()
        cls.draw_format = factories.DrawFormatFactory.create(teams=4)

        factories.StageFactory.reset_sequence()
        cls.stage = factories.StageFactory.create()

        factories.TeamFactory.reset_sequence()
        cls.teams = factories.TeamFactory.create_batch(4, division=cls.stage.division)

    def test_draw_format_name(self):
        self.assertEqual(self.draw_format.name, "Round Robin (3/4 teams)")

    def test_draw_format_text(self):
        self.assertEqual(
            self.draw_format.text,
            "ROUND\n"
            "1: 1 vs 4\n"
            "2: 2 vs 3\n"
            "ROUND\n"
            "3: 1 vs 3\n"
            "4: 4 vs 2\n"
            "ROUND\n"
            "5: 1 vs 2\n"
            "6: 3 vs 4",
        )

    def test_draw_generation_match_form_set_save(self):
        self.assertEqual(Match.objects.count(), 0)

        data0 = {
            "draw_generation_wizard-current_step": "0",
            "0-TOTAL_FORMS": "1",
            "0-INITIAL_FORMS": "1",
            "0-MIN_NUM_FORMS": "0",
            "0-MAX_NUM_FORMS": "1000",
            "0-0-start_date": "2018-8-8",
            "0-0-format": str(self.draw_format.pk),
            "0-0-rounds": "",
            "0-0-offset": "",
        }

        with self.login(self.superuser):
            self.post(
                "admin:fixja:competition:season:division:stage:build",
                self.stage.division.season.competition.pk,
                self.stage.division.season.pk,
                self.stage.division.pk,
                self.stage.pk,
                data=data0,
            )
            self.response_200()

            wizard = self.get_context("wizard")
            forms = wizard["form"]
            total_forms = initial_forms = len(forms)

            data1 = {
                "draw_generation_wizard-current_step": "1",
                "1-TOTAL_FORMS": str(total_forms),
                "1-INITIAL_FORMS": str(initial_forms),
                "1-MIN_NUM_FORMS": "0",
                "1-MAX_NUM_FORMS": "1000",
            }

            for i, form in enumerate(forms):
                data1.update(
                    {
                        "1-%d-round" % i: str(form["round"].initial),
                        "1-%d-home_team" % i: str(form["home_team"].initial or ""),
                        "1-%d-away_team" % i: str(form["away_team"].initial or ""),
                        "1-%d-date" % i: form["date"].initial.strftime("%Y-%m-%d"),
                    }
                )

            self.post(
                "admin:fixja:competition:season:division:stage:build",
                self.stage.division.season.competition.pk,
                self.stage.division.season.pk,
                self.stage.division.pk,
                self.stage.pk,
                data=data1,
            )
            self.response_302()

        self.assertEqual(Match.objects.count(), 6)

        team1, team2, team3, team4 = self.teams
        self.assertQuerySetEqual(
            Match.objects.all(),
            [
                "<Match: 1: %s vs %s>" % (team1, team4),
                "<Match: 1: %s vs %s>" % (team2, team3),
                "<Match: 2: %s vs %s>" % (team1, team3),
                "<Match: 2: %s vs %s>" % (team4, team2),
                "<Match: 3: %s vs %s>" % (team1, team2),
                "<Match: 3: %s vs %s>" % (team3, team4),
            ],
            transform=repr,
        )

    def test_draw_generation_match_form_set_progression_save(self):
        stage = factories.StageFactory.create(
            division=self.stage.division,
            follows=self.stage,
        )

        draw_format = factories.DrawFormatFactory.create(
            teams=4,
            name="Single Elimination",
            text=(
                "ROUND\n"
                "1: P1 vs P4 Semi Final 1\n"
                "2: P2 vs P3 Semi Final 2\n"
                "ROUND\n"
                "3: L1 vs L2 Bronze Medal\n"
                "4: W1 vs W2 Gold Medal\n"
            ),
        )

        self.assertEqual(Match.objects.count(), 0)

        data0 = {
            "draw_generation_wizard-current_step": "0",
            "0-TOTAL_FORMS": "1",
            "0-INITIAL_FORMS": "1",
            "0-MIN_NUM_FORMS": "0",
            "0-MAX_NUM_FORMS": "1000",
            "0-0-start_date": "2018-8-8",
            "0-0-format": str(draw_format.pk),
            "0-0-rounds": "",
            "0-0-offset": "",
        }

        with self.login(self.superuser):
            self.post(
                "admin:fixja:competition:season:division:stage:build",
                stage.division.season.competition.pk,
                stage.division.season.pk,
                stage.division.pk,
                stage.pk,
                data=data0,
            )
            self.response_200()

            wizard = self.get_context("wizard")
            forms = wizard["form"]
            total_forms = initial_forms = len(forms)

            data1 = {
                "draw_generation_wizard-current_step": "1",
                "1-TOTAL_FORMS": str(total_forms),
                "1-INITIAL_FORMS": str(initial_forms),
                "1-MIN_NUM_FORMS": "0",
                "1-MAX_NUM_FORMS": "1000",
            }

            for i, form in enumerate(forms):
                data1.update(
                    {
                        "1-%d-round" % i: str(form["round"].initial),
                        "1-%d-home_team" % i: str(form["home_team"].initial or ""),
                        "1-%d-away_team" % i: str(form["away_team"].initial or ""),
                        "1-%d-date" % i: form["date"].initial.strftime("%Y-%m-%d"),
                    }
                )

            self.post(
                "admin:fixja:competition:season:division:stage:build",
                stage.division.season.competition.pk,
                stage.division.season.pk,
                stage.division.pk,
                stage.pk,
                data=data1,
            )
            self.response_302()

        self.assertCountEqual(
            [(m.get_home_team(), m.get_away_team()) for m in Match.objects.all()],
            [
                ({"title": "1st"}, {"title": "4th"}),
                ({"title": "2nd"}, {"title": "3rd"}),
                ({"title": "Loser Semi Final 1"}, {"title": "Loser Semi Final 2"}),
                ({"title": "Winner Semi Final 1"}, {"title": "Winner Semi Final 2"}),
            ],
        )



import logging
import os.path
from importlib.machinery import ModuleSpec
from importlib.util import module_from_spec
from urllib.parse import urlunparse

from django.conf import settings
from django.contrib.sitemaps.views import sitemap
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.http import HttpResponseForbidden
from django.shortcuts import redirect, render
from django.template import RequestContext, TemplateDoesNotExist
from django.urls import include, path, reverse_lazy
from django.utils.deprecation import MiddlewareMixin
from guardian.conf import settings as guardian_settings
from mptt.utils import tree_item_iterator

from touchtechnology.common.default_settings import (
    SITEMAP_HTTPS_OPTION,
    SITEMAP_ROOT,
)
from touchtechnology.common.models import SitemapNode
from touchtechnology.common.sitemaps import NodeSitemap
from touchtechnology.common.sites import AccountsSite
from touchtechnology.common.views import login
from touchtechnology.content.models import Redirect
from touchtechnology.content.views import dispatch

DEHYDRATED_URLPATTERNS_KEY = "urlpatterns"
DEHYDRATED_URLPATTERNS_TIMEOUT = 600

logger = logging.getLogger(__name__)
protect = AccountsSite(name="protect")


class SitemapNodeMiddleware(MiddlewareMixin):
    def process_request(self, request):
        # When working with multiple tenants, we want to shard the cache for
        # each of them. Use of the version is a nice way to do this as it will
        # prevent collisions while making the API consistent.
        v_kw = {}
        if hasattr(request, "tenant"):
            v_kw.setdefault("version", request.tenant.schema_name)

        # TODO write a cache backend that will do this automatically, and
        #      contribute it back to django-tenant-schemas

        dehydrated = cache.get(DEHYDRATED_URLPATTERNS_KEY, [], **v_kw)

        if not dehydrated:
            logging.getLogger("newrelic.cache").debug("RECALCULATE URLCONF")

            # We need a secret set of account urls so we can bounce the user
            # here if the page is protected. As we haven't yet embedded our
            # url conf (we're in the middle of building it!) this will need
            # to be found using a reverse_lazy below.
            try:
                root = SitemapNode._tree_manager.root_nodes().first()
            except ObjectDoesNotExist:
                root = None
            dehydrated.append(
                {"route": "p/", "site": protect, "kwargs": {"node": root}}
            )

            enabled_nodes = SitemapNode._tree_manager.all()
            related_nodes = enabled_nodes.select_related("content_type")

            def has_disabled_ancestors(st):
                for ancestor in st["ancestors"]:
                    if not ancestor.enabled:
                        return True
                return False

            def get_absolute_url(n, st):
                assert not n.is_root_node()
                offset = 1 if st["ancestors"][0].slug == SITEMAP_ROOT else 0
                paths = [ancestor.slug for ancestor in st["ancestors"][offset:]]
                if paths:
                    return os.path.join(os.path.join(*paths), n.slug)
                return n.slug

            for node, struct in tree_item_iterator(related_nodes, True, lambda x: x):
                # Skip over nodes that they themselves or have disabled ancestors.
                if not node.enabled:
                    logger.debug("%r is disabled, omit from urlconf", node)
                    continue
                if has_disabled_ancestors(struct):
                    logger.debug("%r has disabled ancestor, omit from urlconf", node)
                    continue

                if node.is_root_node() and node.slug == SITEMAP_ROOT:
                    part = ""
                elif node.is_root_node():
                    part = node.slug
                else:
                    part = get_absolute_url(node, struct)

                if part and settings.APPEND_SLASH:
                    part += "/"

                if (
                    node.content_type is not None
                    and node.content_type.model == "placeholder"
                ):
                    try:
                        app = node.object.site(node)
                    except (AttributeError, ImportError, ValueError):
                        logger.exception(
                            "Application is unavailable, disabling this node."
                        )
                        node.disable()
                    else:
                        pattern = {
                            "route": part,
                            "site": app,
                            "kwargs": dict(node=node, **node.kwargs),
                            "name": app.name,
                        }
                        # When nesting applications we need to ensure that any
                        # root url is not clobbered by the patterns of the
                        # parent application. In these cases, force them to the
                        # top of the map.
                        if (
                            node.parent
                            and node.parent.content_type
                            and node.parent.content_type.model == "placeholder"
                        ):
                            dehydrated.insert(0, pattern)
                        else:
                            dehydrated.append(pattern)

                elif node.object_id is None:
                    dehydrated.append(
                        {
                            "route": part,
                            "view": dispatch,
                            "kwargs": dict(node=node, url=part),
                            "name": f"folder_{node.pk}",
                        }
                    )

                else:
                    dehydrated.append(
                        {
                            "route": part,
                            "view": dispatch,
                            "kwargs": dict(page_id=node.object_id, node=node, url=part),
                            "name": f"page_{node.object_id if node.object_id else None}",
                        }
                    )

            cache.set(
                DEHYDRATED_URLPATTERNS_KEY,
                dehydrated,
                timeout=DEHYDRATED_URLPATTERNS_TIMEOUT,
                **v_kw,
            )

        # Always start with the project wide ROOT_URLCONF and add our sitemap.xml view
        urlpatterns = [
            path(
                "sitemap.xml",
                sitemap,
                {"sitemaps": {"nodes": NodeSitemap}},
                name="sitemap",
            ),
            path("", include(settings.ROOT_URLCONF)),
        ]

        # Construct the cache of url pattern definitions. We are not keeping
        # the actual patterns, because pickling is problematic for the .url
        # instancemethod - instead we keep the skeleton and build it on the
        # fly from cache... rehydrating it ;)
        for node in dehydrated:
            try:
                pattern = path(
                    node["route"],
                    node["view"],
                    node["kwargs"],
                    name=node.get("name"),
                )
            except KeyError:
                pattern = path(
                    node["route"],
                    node["site"].urls,
                    node["kwargs"],
                    name=node["site"].name,
                )
            urlpatterns.append(pattern)

        # Create a new module on the fly and attach the rehydrated urlpatterns
        dynamic_urls = module_from_spec(ModuleSpec("dynamic_urls", None))
        dynamic_urls.urlpatterns = urlpatterns

        # Attach the module to the request
        request.urlconf = dynamic_urls

    def process_view(self, request, view_func, view_args, view_kwargs):
        node = view_kwargs.get("node")

        if node:
            required = node.restrict_to_groups.all()

            if required:
                next = node.get_absolute_url()
                to = reverse_lazy("accounts:login")

                # An anonymous user will never be a member of a group, so make
                # them go off and be authenticated.
                if request.user.is_anonymous():
                    return login(request, to=to, next=next)

                # A user who is not a member of a suitable group should get a
                # 403 page.
                groups = request.user.groups.all()
                if set(required).difference(groups):
                    if guardian_settings.RENDER_403:
                        try:
                            response = render(
                                request,
                                guardian_settings.TEMPLATE_403,
                                {},
                                RequestContext(request),
                            )
                            response.status_code = 403
                            return response
                        except TemplateDoesNotExist as e:
                            if settings.DEBUG:
                                raise e
                    elif guardian_settings.RAISE_403:
                        raise PermissionDenied
                    return HttpResponseForbidden()

        if (
            SITEMAP_HTTPS_OPTION
            and node
            and node.require_https
            and not request.is_secure()
        ):
            host = request.META.get("HTTP_HOST")
            path = request.META.get("PATH_INFO")
            redirect_to = urlunparse(("https", host, path, "", "", ""))
            return redirect(redirect_to)


def redirect_middleware(get_response):
    def middleware(request):
        objects = Redirect.objects.filter(source_url__exact=request.path)
        if objects.exists():
            instance = objects.first()
            return redirect(instance.destination_url, instance.permanent)
        return get_response(request)

    return middleware



import datetime
import logging

from dateutil.parser import parse as parse_datetime
from dateutil.relativedelta import relativedelta
from django.http import Http404
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.functional import wraps
from django.views.decorators.http import last_modified

from touchtechnology.common.models import SitemapNode
from touchtechnology.news.models import Article, Category

logger = logging.getLogger(__name__)


@method_decorator
def date_view(f, *a, **kw):
    @wraps(f)
    def _decorated(*args, **kwargs):
        year = kwargs.pop("year")
        month = kwargs.pop("month", "jan")
        day = kwargs.pop("day", 1)

        try:
            value = timezone.make_aware(
                parse_datetime(f"{year}-{month}-{day}"),
                datetime.timezone.utc,
            )
        except ValueError as exc:
            logger.exception("invalid date value in path %s", args[0].path)
            raise Http404(str(exc))

        kwargs["date"] = value

        # So we can run an optimal query for finding the last modified time for
        # a view in other decorators, pass in the "delta" that should be used
        if day is None:
            kwargs["delta"] = "months"
        if month is None:
            kwargs["delta"] = "years"

        return f(*args, **kwargs)

    return _decorated


@method_decorator
@last_modified
def news_last_modified(request, **kwargs):
    last_modified_datetimes = []
    try:
        last_modified_datetimes.append(
            Article.objects.live().latest("last_modified").last_modified
        )
    except Article.DoesNotExist:
        ...
    try:
        last_modified_datetimes.append(
            Category.objects.latest("last_modified").last_modified
        )
    except Category.DoesNotExist:
        ...
    try:
        last_modified_datetimes.append(
            SitemapNode.objects.latest("last_modified").last_modified
        )
    except SitemapNode.DoesNotExist:
        ...
    return max(last_modified_datetimes)


@method_decorator
@last_modified
def last_modified_article(request, **kwargs):
    queryset = Article.objects.live()

    date = kwargs.get("date")
    if date is not None:
        delta = {kwargs.get("delta", "days"): 1}
        date_range = (date, date + relativedelta(**delta))
        queryset = queryset.filter(published__range=date_range)

    slug = kwargs.get("slug")
    if slug is not None:
        queryset = queryset.filter(slug=slug)

    if not queryset:
        return
    return queryset.latest("last_modified").last_modified



from django.test.utils import override_settings
from test_plus import TestCase


@override_settings(ROOT_URLCONF="example_app.urls")
class DateTimeFieldTest(TestCase):
    def test_validate_form_output_tz(self):
        self.get("datetime:index")
        self.assertResponseContains(
            '<option value="Australia/Sydney">GMT+10:00 Australia/Sydney</option>'
        )

    def test_submit_date_time_valid(self):
        data = {
            "form-datetime_0": "2014-1-1",
            "form-datetime_1": "0:00",
            "form-datetime_2": "Australia/Sydney",
            "formset-TOTAL_FORMS": "0",
            "formset-INITIAL_FORMS": "0",
            "formset-MAX_NUM_FORMS": "1000",
        }
        self.post("datetime:index", data=data)
        self.response_302()

    def test_submit_date_time_blank(self):
        data = {
            "form-datetime_0": "",
            "form-datetime_1": "",
            "form-datetime_2": "",
            "formset-TOTAL_FORMS": "0",
            "formset-INITIAL_FORMS": "0",
            "formset-MAX_NUM_FORMS": "1000",
        }
        self.post("datetime:index", data=data)
        self.response_302()



from rest_framework import serializers
from rest_framework_nested.serializers import NestedHyperlinkedModelSerializer

from tournamentcontrol.competition import models

from .club import ClubSerializer
from .season import PlaceSerializer
from .viewsets import SlugViewSet


class ListMatchSerializer(serializers.ModelSerializer):
    round = serializers.SerializerMethodField(read_only=True)
    home_team = serializers.SerializerMethodField(read_only=True)
    away_team = serializers.SerializerMethodField(read_only=True)
    play_at = PlaceSerializer(read_only=True)

    class Meta:
        model = models.Match
        fields = (
            "id",
            "uuid",
            "round",
            "date",
            "time",
            "datetime",
            "is_bye",
            "is_washout",
            "home_team",
            "home_team_score",
            "away_team",
            "away_team_score",
            "referees",
            "videos",
            "play_at",
        )

    def get_round(self, obj):
        return obj.label or "Round {}".format(obj.round)

    def _get_team(self, obj, home_or_away):
        team = getattr(obj, f"get_{home_or_away}_team_plain")()
        if isinstance(team, (models.Team, models.ByeTeam)):
            return team.pk
        return team

    def get_home_team(self, obj):
        return self._get_team(obj, "home")

    def get_away_team(self, obj):
        return self._get_team(obj, "away")


class ListStageSerializer(NestedHyperlinkedModelSerializer):
    parent_lookup_kwargs = {
        "competition_slug": "division__season__competition__slug",
        "season_slug": "division__season__slug",
        "division_slug": "division__slug",
    }

    matches = ListMatchSerializer(many=True, read_only=True)

    class Meta:
        model = models.Stage
        fields = ("title", "slug", "url", "matches")
        extra_kwargs = {"url": {"lookup_field": "slug", "view_name": "v1:stage-detail"}}


class ListTeamSerializer(serializers.ModelSerializer):

    club = ClubSerializer(read_only=True)

    class Meta:
        model = models.Team
        fields = ("id", "title", "slug", "club")


class ListDivisionSerializer(NestedHyperlinkedModelSerializer):
    parent_lookup_kwargs = {
        "competition_slug": "season__competition__slug",
        "season_slug": "season__slug",
    }

    class Meta:
        model = models.Division
        fields = ("title", "slug", "url")
        extra_kwargs = {
            "url": {"lookup_field": "slug", "view_name": "v1:division-detail"},
            "season": {"lookup_field": "slug"},
        }


class DivisionSerializer(ListDivisionSerializer):
    teams = ListTeamSerializer(many=True, read_only=True)
    stages = ListStageSerializer(many=True, read_only=True)

    class Meta(ListDivisionSerializer.Meta):
        fields = ("title", "slug", "url", "teams", "stages")


class DivisionViewSet(SlugViewSet):
    serializer_class = DivisionSerializer
    list_serializer_class = ListDivisionSerializer

    def get_queryset(self):
        return (
            models.Division.objects.filter(
                season__slug=self.kwargs["season_slug"],
                season__competition__slug=self.kwargs["competition_slug"],
            )
            .select_related("season__competition")
            .prefetch_related(
                "teams__club",
                "stages__matches__home_team",
                "stages__matches__away_team",
            )
        )



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("news", "0007_article_copy_field_data"),
    ]

    operations = [
        migrations.DeleteModel(
            name="ArticleContent",
        ),
    ]



import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("competition", "0046_alter_match_live_stream_bind"),
    ]

    operations = [
        migrations.AddField(
            model_name="season",
            name="live_stream_client_id",
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name="season",
            name="live_stream_client_secret",
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name="season",
            name="live_stream_project_id",
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name="season",
            name="live_stream_refresh_token",
            field=models.CharField(max_length=200, null=True),
        ),
        migrations.AddField(
            model_name="season",
            name="live_stream_scopes",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=200), null=True, size=None
            ),
        ),
        migrations.AddField(
            model_name="season",
            name="live_stream_token",
            field=models.CharField(max_length=1000, null=True),
        ),
        migrations.AddField(
            model_name="season",
            name="live_stream_token_uri",
            field=models.URLField(null=True),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations

import tournamentcontrol.competition.models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0010_remove_match_video_approved_by"),
    ]

    operations = [
        migrations.AlterField(
            model_name="club",
            name="twitter",
            field=tournamentcontrol.competition.models.TwitterField(
                help_text='Official Twitter name for use in social "mentions"',
                max_length=50,
                blank=True,
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="division",
            name="points_formula",
            field=tournamentcontrol.competition.models.LadderPointsField(
                null=True,
                verbose_name="Points system",
                blank=True,
            ),
            preserve_default=True,
        ),
    ]



from tournamentcontrol.competition.rest.v1._routers import (
    competition_router,
    division_router,
    router,
    season_router,
)

urlpatterns = (
    router.urls + competition_router.urls + season_router.urls + division_router.urls
)



from django import forms
from django.contrib.auth import get_user_model
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from touchtechnology.common.forms import RedefineModelForm


class ProfileForm(RedefineModelForm):
    def clean_email(self):
        UserModel = get_user_model()
        email = self.cleaned_data.get("email")
        other_users = UserModel.objects.exclude(pk=self.instance.pk)
        if other_users.filter(email__iexact=email):
            raise forms.ValidationError(
                _(
                    "This email address is already in "
                    "use. Please supply a different "
                    "email address."
                )
            )
        return email

    class Meta:
        model = get_user_model()
        fields = (
            "first_name",
            "last_name",
            "email",
        )
        redefine = (
            ("first_name", {"required": True}),
            ("last_name", {"required": True}),
            (
                "email",
                {
                    "help_text": mark_safe(
                        _(
                            "Your email address is also your "
                            "username.<br />Please ensure you "
                            "enter it correctly."
                        )
                    ),
                    "required": True,
                },
            ),
        )



from django.apps import apps
from django.contrib.contenttypes.fields import GenericForeignKey
from django.db import transaction
from django.db.models import Model, OneToOneField


@transaction.atomic
def merge_model_objects(primary_object, alias_objects=[], keep_old=False):
    """
    Use this function to merge model objects (i.e. Users, Organizations, Polls,
    etc.) and migrate all of the related fields from the alias objects to the
    primary object.

    Usage:
    from django.contrib.auth.models import User
    primary_user = User.objects.get(email='good_email@example.com')
    duplicate_user = User.objects.get(email='good_email+duplicate@example.com')
    merge_model_objects(primary_user, duplicate_user)
    """
    if not isinstance(alias_objects, list):
        alias_objects = [alias_objects]

    # check that all aliases are the same class as primary one and that
    # they are subclass of model
    primary_class = primary_object.__class__

    if not issubclass(primary_class, Model):
        raise TypeError("Only django.db.models.Model subclasses can be merged")

    for alias_object in alias_objects:
        if not isinstance(alias_object, primary_class):
            raise TypeError("Only models of same class can be merged")

    # Get a list of all GenericForeignKeys in all models
    # TODO: this is a bit of a hack, since the generics framework should provide a similar
    # method to the ForeignKey field for accessing the generic related fields.
    generic_fields = [
        field
        for app in apps.get_app_configs()
        for model in app.get_models()
        for field_name, field in model.__dict__.items()
        if isinstance(field, GenericForeignKey)
    ]

    blank_local_fields = {
        field.attname
        for field in primary_object._meta.local_fields
        if getattr(primary_object, field.attname) in {None, ""}
    }

    # Loop through all alias objects and migrate their data to the primary object.
    for alias_object in alias_objects:
        # Migrate all foreign key references from alias object to primary object.
        for related_object in [
            f
            for f in alias_object._meta.get_fields()
            if (f.one_to_many or f.one_to_one) and f.auto_created and not f.concrete
        ]:
            # The variable name on the alias_object model.
            alias_varname = related_object.get_accessor_name()
            if isinstance(related_object.field, OneToOneField):
                break
            # The variable name on the related model.
            obj_varname = related_object.field.name
            related_objects = getattr(alias_object, alias_varname)
            for obj in related_objects.all():
                setattr(obj, obj_varname, primary_object)
                obj.save()

        # Migrate all many to many references from alias object to primary object.
        for related_many_object in [
            f
            for f in alias_object._meta.get_fields(include_hidden=True)
            if f.many_to_many and f.auto_created
        ]:
            alias_varname = related_many_object.get_accessor_name()
            obj_varname = related_many_object.field.name

            if alias_varname is not None:
                # standard case
                related_many_objects = getattr(alias_object, alias_varname).all()
            else:
                # special case, symmetrical relation, no reverse accessor
                related_many_objects = getattr(alias_object, obj_varname).all()
            for obj in related_many_objects.all():
                getattr(obj, obj_varname).remove(alias_object)
                getattr(obj, obj_varname).add(primary_object)

        # Migrate all generic foreign key references from alias object to primary object.
        for field in generic_fields:
            filter_kwargs = {}
            filter_kwargs[field.fk_field] = alias_object._get_pk_val()
            filter_kwargs[field.ct_field] = field.get_content_type(alias_object)
            for generic_related_object in field.model.objects.filter(**filter_kwargs):
                setattr(generic_related_object, field.name, primary_object)
                generic_related_object.save()

        # Try to fill all missing values in primary object by values of duplicates
        filled_up = set()
        for field_name in blank_local_fields:
            val = getattr(alias_object, field_name)
            if val not in [None, ""]:
                setattr(primary_object, field_name, val)
                filled_up.add(field_name)
        blank_local_fields -= filled_up

        if not keep_old:
            alias_object.delete()
    primary_object.save()
    return primary_object



# coding=UTF-8

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from touchtechnology.common.forms.auth import permissionformset_factory
from touchtechnology.common.forms.fields import EmailField, HTMLField
from touchtechnology.common.models import SitemapNode
from touchtechnology.common.tests import factories


class CustomFormField(TestCase):
    def test_email_field(self):
        self.assertFieldOutput(
            EmailField,
            {"a@a.com": "a@a.com", "B@B.COM": "b@b.com"},
            {"aaa": ["Enter a valid email address."]},
        )
        self.assertFieldOutput(
            EmailField,
            {"a@a.com": "a@a.com", "B@B.COM": "B@B.COM"},
            {"aaa": ["Enter a valid email address."]},
            (),
            {"lowercase": False},
        )

    def test_html_field(self):
        valid = {
            '<a href="http://www.example.com/">Example</a>': '<a href="http://www.example.com/">Example</a>',
            "Penn\u00a0& Teller": "Penn&nbsp;& Teller",
            "sauté": "saut&eacute;",
        }
        self.assertFieldOutput(HTMLField, valid, {})

    maxDiff = None


class TestPermissionFormSet(TestCase):
    def setUp(self):
        self.UserModel = get_user_model()
        self.queryset = Permission.objects.filter(
            content_type=ContentType.objects.get_for_model(SitemapNode)
        ).exclude(codename__startswith="add_")
        self.instance = SitemapNode.objects.create(title="permissions")
        self.staff = factories.UserFactory.create(is_staff=True, is_superuser=True)
        self.regular = factories.UserFactory.create()

    def test_staff_only(self):
        formset_class = permissionformset_factory(SitemapNode, staff_only=True)
        formset = formset_class(queryset=self.queryset, instance=self.instance)
        self.assertQuerySetEqual(
            formset.forms[0].fields["users"].queryset,
            (repr(o) for o in self.UserModel.objects.filter(is_staff=True)),
            transform=repr,
        )

    def test_all_users(self):
        formset_class = permissionformset_factory(SitemapNode, staff_only=False)
        formset = formset_class(queryset=self.queryset, instance=self.instance)
        self.assertQuerySetEqual(
            formset.forms[0].fields["users"].queryset.order_by("pk"),
            (repr(o) for o in self.UserModel.objects.order_by("pk")),
            transform=repr,
        )

    def test_user_widget_checkbox_lte(self):
        "Less than equal to 5 users, should be iCheckboxSelectMultiple widget"
        formset_class = permissionformset_factory(
            SitemapNode, staff_only=False, max_checkboxes=5
        )
        formset = formset_class(queryset=self.queryset, instance=self.instance)
        self.assertEqual(
            formset.forms[0].fields["users"].queryset.count(),
            3,  # need to account for the AnonymousUser that guardian creates
        )

    def test_user_widget_checkbox_eq(self):
        "Equal to 3 users, should be iCheckboxSelectMultiple widget"
        formset_class = permissionformset_factory(
            SitemapNode, staff_only=False, max_checkboxes=3
        )
        formset = formset_class(queryset=self.queryset, instance=self.instance)
        self.assertEqual(
            formset.forms[0].fields["users"].queryset.count(),
            3,  # need to account for the AnonymousUser that guardian creates
        )

    def test_user_widget_select2(self):
        "More than 1 user, should not be iCheckboxSelectMultiple widget"
        formset_class = permissionformset_factory(
            SitemapNode, staff_only=False, max_checkboxes=1
        )
        formset = formset_class(queryset=self.queryset, instance=self.instance)
        self.assertEqual(
            formset.forms[0].fields["users"].queryset.count(),
            3,  # need to account for the AnonymousUser that guardian creates
        )



# -*- coding: utf-8 -*-

import django.db.models.deletion
from django.db import migrations, models

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Club",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "short_title",
                    models.CharField(
                        help_text="This is used in navigation menus instead of the longer title value.",
                        max_length=100,
                        verbose_name="Short title",
                        blank=True,
                    ),
                ),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "slug_locked",
                    models.BooleanField(default=False, verbose_name="Slug locked"),
                ),
                ("email", models.EmailField(max_length=255, blank=True)),
                ("website", models.URLField(max_length=255, blank=True)),
                (
                    "twitter",
                    models.CharField(
                        help_text='Official Twitter name for use in social "mentions"',
                        max_length=50,
                        blank=True,
                    ),
                ),
                (
                    "abbreviation",
                    models.CharField(
                        help_text="Optional 3-letter abbreviation to be used on scoreboards.",
                        max_length=3,
                        blank=True,
                    ),
                ),
            ],
            options={
                "ordering": ("title",),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="ClubAssociation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "club",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="staff",
                        to="competition.Club",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": ("person__last_name", "person__first_name"),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="ClubRole",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=50)),
            ],
            options={
                "ordering": ("name",),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Competition",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "short_title",
                    models.CharField(
                        help_text="This is used in navigation menus instead of the longer title value.",
                        max_length=100,
                        verbose_name="Short title",
                        blank=True,
                    ),
                ),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "slug_locked",
                    models.BooleanField(default=False, verbose_name="Slug locked"),
                ),
                ("order", models.PositiveIntegerField(default=1)),
                ("copy", touchtechnology.common.db.models.HTMLField(blank=True)),
                ("enabled", models.BooleanField(default=True)),
                (
                    "clubs",
                    touchtechnology.common.db.models.ManyToManyField(
                        related_name="competitions",
                        null=True,
                        to="competition.Club",
                        blank=True,
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Division",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "short_title",
                    models.CharField(
                        help_text="This is used in navigation menus instead of the longer title value.",
                        max_length=100,
                        verbose_name="Short title",
                        blank=True,
                    ),
                ),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "slug_locked",
                    models.BooleanField(default=False, verbose_name="Slug locked"),
                ),
                ("order", models.PositiveIntegerField(default=1)),
                (
                    "points_formula",
                    models.TextField(
                        null=True, verbose_name="Points system", blank=True
                    ),
                ),
                ("bonus_points_formula", models.TextField(null=True, blank=True)),
                ("forfeit_for_score", models.SmallIntegerField(null=True)),
                ("forfeit_against_score", models.SmallIntegerField(null=True)),
                ("include_forfeits_in_played", models.BooleanField(default=False)),
                (
                    "games_per_day",
                    models.SmallIntegerField(
                        help_text="\n        In Tournament mode, specify how many matches per day should be scheduled\n        by the automatic draw generator.",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="DivisionExclusionDate",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("date", touchtechnology.common.db.models.DateField()),
                (
                    "division",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="exclusions",
                        to="competition.Division",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "verbose_name": "Exclusion Date",
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="DrawFormat",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=50)),
                ("text", models.TextField(verbose_name="Formula")),
                ("teams", models.PositiveIntegerField(null=True, blank=True)),
                ("is_final", models.BooleanField(default=False)),
            ],
            options={
                "ordering": ("is_final", "teams", "name"),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="LadderEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("played", models.SmallIntegerField(default=0)),
                ("win", models.SmallIntegerField(default=0)),
                ("loss", models.SmallIntegerField(default=0)),
                ("draw", models.SmallIntegerField(default=0)),
                ("bye", models.SmallIntegerField(default=0)),
                ("forfeit_for", models.SmallIntegerField(default=0)),
                ("forfeit_against", models.SmallIntegerField(default=0)),
                ("score_for", models.SmallIntegerField(default=0)),
                ("score_against", models.SmallIntegerField(default=0)),
                ("bonus_points", models.SmallIntegerField(default=0)),
                (
                    "points",
                    models.DecimalField(default=0, max_digits=6, decimal_places=3),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="LadderSummary",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("played", models.SmallIntegerField(default=0)),
                ("win", models.SmallIntegerField(default=0)),
                ("loss", models.SmallIntegerField(default=0)),
                ("draw", models.SmallIntegerField(default=0)),
                ("bye", models.SmallIntegerField(default=0)),
                ("forfeit_for", models.SmallIntegerField(default=0)),
                ("forfeit_against", models.SmallIntegerField(default=0)),
                ("score_for", models.SmallIntegerField(default=0)),
                ("score_against", models.SmallIntegerField(default=0)),
                ("bonus_points", models.SmallIntegerField(default=0)),
                (
                    "points",
                    models.DecimalField(default=0, max_digits=6, decimal_places=3),
                ),
                (
                    "difference",
                    models.DecimalField(default=0, max_digits=6, decimal_places=3),
                ),
                (
                    "percentage",
                    models.DecimalField(null=True, max_digits=10, decimal_places=2),
                ),
            ],
            options={
                "ordering": (
                    "stage",
                    "-points",
                    "-difference",
                    "-percentage",
                    "team__title",
                ),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Match",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("label", models.CharField(max_length=100, null=True, blank=True)),
                (
                    "home_team_eval",
                    models.CharField(max_length=10, null=True, blank=True),
                ),
                (
                    "away_team_eval",
                    models.CharField(max_length=10, null=True, blank=True),
                ),
                ("evaluated", models.NullBooleanField()),
                ("is_washout", models.BooleanField(default=False)),
                (
                    "date",
                    touchtechnology.common.db.models.DateField(null=True, blank=True),
                ),
                (
                    "time",
                    touchtechnology.common.db.models.TimeField(null=True, blank=True),
                ),
                (
                    "datetime",
                    touchtechnology.common.db.models.DateTimeField(
                        null=True, blank=True
                    ),
                ),
                ("home_team_score", models.IntegerField(null=True, blank=True)),
                ("away_team_score", models.IntegerField(null=True, blank=True)),
                ("is_bye", models.BooleanField(default=False)),
                ("bye_processed", models.BooleanField(default=False)),
                ("is_forfeit", models.BooleanField(default=False)),
                ("round", models.IntegerField(null=True, blank=True)),
                ("include_in_ladder", models.BooleanField(default=True)),
            ],
            options={
                "ordering": (
                    "date",
                    "stage",
                    "round",
                    "is_bye",
                    "time",
                    "play_at__ground__order",
                    "id",
                ),
                "verbose_name_plural": "matches",
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("first_name", models.CharField(max_length=30)),
                ("last_name", models.CharField(max_length=30)),
                (
                    "gender",
                    models.CharField(
                        max_length=1, choices=[(b"M", "Male"), (b"F", "Female")]
                    ),
                ),
                (
                    "date_of_birth",
                    touchtechnology.common.db.models.DateField(null=True, blank=True),
                ),
                ("email", models.EmailField(max_length=75, blank=True)),
                ("home_phone", models.CharField(max_length=30, blank=True)),
                ("work_phone", models.CharField(max_length=30, blank=True)),
                ("mobile_phone", models.CharField(max_length=30, blank=True)),
                (
                    "club",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="members",
                        to="competition.Club",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": ("last_name", "first_name"),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Place",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "short_title",
                    models.CharField(
                        help_text="This is used in navigation menus instead of the longer title value.",
                        max_length=100,
                        verbose_name="Short title",
                        blank=True,
                    ),
                ),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "slug_locked",
                    models.BooleanField(default=False, verbose_name="Slug locked"),
                ),
                ("order", models.PositiveIntegerField(default=1)),
                (
                    "abbreviation",
                    models.CharField(max_length=20, null=True, blank=True),
                ),
                (
                    "latlng",
                    touchtechnology.common.db.models.LocationField(max_length=100),
                ),
                ("timezone", models.CharField(blank=True, max_length=50, null=True)),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Ground",
            fields=[
                (
                    "place_ptr",
                    models.OneToOneField(
                        parent_link=True,
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        to="competition.Place",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
            bases=("competition.place",),
        ),
        migrations.CreateModel(
            name="Season",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "short_title",
                    models.CharField(
                        help_text="This is used in navigation menus instead of the longer title value.",
                        max_length=100,
                        verbose_name="Short title",
                        blank=True,
                    ),
                ),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "slug_locked",
                    models.BooleanField(default=False, verbose_name="Slug locked"),
                ),
                ("order", models.PositiveIntegerField(default=1)),
                (
                    "hashtag",
                    models.CharField(
                        help_text="Your official tag for social media promotions, excluding the # symbol.",
                        max_length=30,
                        null=True,
                        verbose_name="Hash Tag",
                        blank=True,
                    ),
                ),
                ("enabled", models.BooleanField(default=True)),
                (
                    "start_date",
                    touchtechnology.common.db.models.DateField(null=True, blank=True),
                ),
                (
                    "mode",
                    models.IntegerField(
                        default=2,
                        help_text="Used by the draw wizard to help you set your match dates & times automatically.",
                        choices=[(2, "Season"), (3, "Tournament")],
                    ),
                ),
                (
                    "statistics",
                    models.BooleanField(
                        default=True,
                        help_text="Set to No if you do not wish to keep scoring or most valuable player statistics.",
                    ),
                ),
                (
                    "mvp_results_public",
                    touchtechnology.common.db.models.DateTimeField(
                        help_text="Set to prevent the front-end site revealing MVP points before the specified time.",
                        null=True,
                        blank=True,
                    ),
                ),
                (
                    "complete",
                    models.BooleanField(
                        default=False,
                        help_text="Set to indicate this season is in the past. Optimises progression calculations.",
                    ),
                ),
                ("timezone", models.CharField(blank=True, max_length=50, null=True)),
                (
                    "competition",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="seasons",
                        to="competition.Competition",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="SeasonAssociation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "club",
                    models.ForeignKey(
                        related_name="officials",
                        to="competition.Club",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
                (
                    "person",
                    touchtechnology.common.db.models.ForeignKey(
                        to="competition.Person",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
                (
                    "roles",
                    touchtechnology.common.db.models.ManyToManyField(
                        to="competition.ClubRole"
                    ),
                ),
                (
                    "season",
                    models.ForeignKey(
                        related_name="officials",
                        to="competition.Season",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": (
                    "club",
                    "season",
                    "person__last_name",
                    "person__first_name",
                ),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="SeasonExclusionDate",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("date", touchtechnology.common.db.models.DateField()),
                (
                    "season",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="exclusions",
                        to="competition.Season",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "verbose_name": "Exclusion Date",
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="SeasonMatchTime",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("start", touchtechnology.common.db.models.TimeField()),
                ("interval", models.IntegerField()),
                ("count", models.IntegerField()),
                (
                    "start_date",
                    touchtechnology.common.db.models.DateField(
                        null=True, verbose_name="From", blank=True
                    ),
                ),
                (
                    "end_date",
                    touchtechnology.common.db.models.DateField(
                        null=True, verbose_name="Until", blank=True
                    ),
                ),
                (
                    "season",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="timeslots",
                        to="competition.Season",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="SimpleScoreMatchStatistic",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("number", models.IntegerField(null=True, blank=True)),
                ("played", models.SmallIntegerField(default=0)),
                ("points", models.SmallIntegerField(default=0)),
                ("mvp", models.SmallIntegerField(default=0)),
                (
                    "match",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="statistics",
                        to="competition.Match",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
                (
                    "player",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="statistics",
                        to="competition.Person",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": ("match", "number"),
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Stage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "short_title",
                    models.CharField(
                        help_text="This is used in navigation menus instead of the longer title value.",
                        max_length=100,
                        verbose_name="Short title",
                        blank=True,
                    ),
                ),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "slug_locked",
                    models.BooleanField(default=False, verbose_name="Slug locked"),
                ),
                ("order", models.PositiveIntegerField(default=1)),
                (
                    "keep_ladder",
                    models.BooleanField(
                        default=True,
                        help_text="\n        Set this to <b>No</b> if this stage does not need to keep a competition ladder.<br />\n        Usually set to No for a Final Series or a Knockout stage.\n        ",
                        verbose_name="Keep a ladder",
                    ),
                ),
                (
                    "scale_group_points",
                    models.BooleanField(
                        default=False,
                        help_text="\n        In stages with multiple pools, adjust points in the smaller groups to\n        compensate for the reduced opportunity to score points.<br />\n        You <strong>should</strong> also set 0 points for Bye matches.",
                    ),
                ),
                (
                    "carry_ladder",
                    models.BooleanField(
                        default=False,
                        help_text="\n        Set this to <b>Yes</b> if this stage should carry over values from the previous stage.\n        ",
                        verbose_name="Carry over points",
                    ),
                ),
                (
                    "keep_mvp",
                    models.BooleanField(
                        default=True,
                        help_text="\n        Set this to <b>No</b> if this stage does not need to keep track of MVP points.<br />\n        Usually set to No for a Final Series.\n        ",
                        verbose_name="Keep MVP stats",
                    ),
                ),
                (
                    "division",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="stages",
                        to="competition.Division",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
                (
                    "follows",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="preceeds",
                        on_delete=django.db.models.deletion.SET_NULL,
                        blank=True,
                        to="competition.Stage",
                        help_text="\n        When progressing teams into this stage, which earlier stage should be used for\n        determining positions.<br />\n        Default is the immediately preceeding stage.\n        ",
                        null=True,
                    ),
                ),
                (
                    "matches_needing_printing",
                    touchtechnology.common.db.models.ManyToManyField(
                        related_name="to_be_printed",
                        to="competition.Match",
                        blank=True,
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="StageGroup",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "short_title",
                    models.CharField(
                        help_text="This is used in navigation menus instead of the longer title value.",
                        max_length=100,
                        verbose_name="Short title",
                        blank=True,
                    ),
                ),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "slug_locked",
                    models.BooleanField(default=False, verbose_name="Slug locked"),
                ),
                ("order", models.PositiveIntegerField(default=1)),
                (
                    "carry_ladder",
                    models.BooleanField(
                        default=False,
                        help_text="\n        Set this to <b>Yes</b> if the ladder for this pool should carry over values from\n        the previous stage.<br />\n        Will only apply for matches played against teams that are now in this group.\n        ",
                        verbose_name="Carry over points",
                    ),
                ),
                (
                    "stage",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="pools",
                        to="competition.Stage",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
                "verbose_name": "Pool",
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Team",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "short_title",
                    models.CharField(
                        help_text="This is used in navigation menus instead of the longer title value.",
                        max_length=100,
                        verbose_name="Short title",
                        blank=True,
                    ),
                ),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "slug_locked",
                    models.BooleanField(default=False, verbose_name="Slug locked"),
                ),
                ("order", models.PositiveIntegerField(default=1)),
                (
                    "names_locked",
                    models.BooleanField(
                        default=False,
                        help_text="\n        When the team name is locked, the team manager will not be able to change their team name.<br />\n        As a tournament manager you can always change the names.",
                    ),
                ),
                (
                    "timeslots_after",
                    touchtechnology.common.db.models.TimeField(
                        help_text="Specify the earliest time that this team can play. Leave blank for no preference.",
                        null=True,
                        verbose_name="Start after",
                        blank=True,
                    ),
                ),
                (
                    "timeslots_before",
                    touchtechnology.common.db.models.TimeField(
                        help_text="Specify the latest time that this team can play. Leave blank for no preference.",
                        null=True,
                        verbose_name="Start before",
                        blank=True,
                    ),
                ),
                (
                    "club",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="teams",
                        on_delete=django.db.models.deletion.SET_NULL,
                        blank=True,
                        to="competition.Club",
                        null=True,
                    ),
                ),
                (
                    "division",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="teams",
                        blank=True,
                        to="competition.Division",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
                (
                    "stage_group",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="teams",
                        on_delete=django.db.models.deletion.SET_NULL,
                        verbose_name="Pool",
                        blank=True,
                        to="competition.StageGroup",
                        null=True,
                    ),
                ),
                (
                    "team_clashes",
                    touchtechnology.common.db.models.ManyToManyField(
                        help_text="Select any teams that must not play at the same time.",
                        related_name="team_clashes_rel_+",
                        verbose_name="Don't clash",
                        to="competition.Team",
                        blank=True,
                    ),
                ),
            ],
            options={
                "ordering": ("division__order", "stage_group__order", "order"),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="TeamAssociation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("number", models.IntegerField(null=True, blank=True)),
                ("is_player", models.BooleanField(default=True)),
                (
                    "person",
                    touchtechnology.common.db.models.ForeignKey(
                        to="competition.Person",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": (
                    "-is_player",
                    "number",
                    "person__last_name",
                    "person__first_name",
                ),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="TeamRole",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=50)),
                (
                    "competition",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="team_roles",
                        to="competition.Competition",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": ("name",),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="UndecidedTeam",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("formula", models.CharField(max_length=20, blank=True)),
                ("label", models.CharField(max_length=30, blank=True)),
                (
                    "stage",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="undecided_teams",
                        to="competition.Stage",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
                (
                    "stage_group",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="undecided_teams",
                        on_delete=django.db.models.deletion.SET_NULL,
                        verbose_name="Pool",
                        blank=True,
                        to="competition.StageGroup",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ("stage_group", "formula"),
                "verbose_name": "team",
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Venue",
            fields=[
                (
                    "place_ptr",
                    models.OneToOneField(
                        parent_link=True,
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        to="competition.Place",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
                (
                    "season",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="venues",
                        to="competition.Season",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
            bases=("competition.place",),
        ),
        migrations.AddField(
            model_name="teamassociation",
            name="roles",
            field=touchtechnology.common.db.models.ManyToManyField(
                to="competition.TeamRole", blank=True
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="teamassociation",
            name="team",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="people",
                to="competition.Team",
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name="teamassociation",
            unique_together=set([("team", "person")]),
        ),
        migrations.AlterUniqueTogether(
            name="team",
            unique_together=set([("title", "division")]),
        ),
        migrations.AlterUniqueTogether(
            name="stagegroup",
            unique_together=set([("stage", "order")]),
        ),
        migrations.AlterUniqueTogether(
            name="stage",
            unique_together=set([("division", "slug")]),
        ),
        migrations.AlterUniqueTogether(
            name="seasonexclusiondate",
            unique_together=set([("season", "date")]),
        ),
        migrations.AlterUniqueTogether(
            name="seasonassociation",
            unique_together=set([("season", "person")]),
        ),
        migrations.AddField(
            model_name="match",
            name="away_team",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="away_games",
                blank=True,
                to="competition.Team",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="match",
            name="away_team_eval_related",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="+away_team_eval",
                blank=True,
                to="competition.Match",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="match",
            name="away_team_undecided",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="away_games",
                blank=True,
                to="competition.UndecidedTeam",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="match",
            name="forfeit_winner",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="+team",
                blank=True,
                to="competition.Team",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="match",
            name="home_team",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="home_games",
                blank=True,
                to="competition.Team",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="match",
            name="home_team_eval_related",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="+home_team_eval",
                blank=True,
                to="competition.Match",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="match",
            name="home_team_undecided",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="home_games",
                blank=True,
                to="competition.UndecidedTeam",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="match",
            name="play_at",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.SET_NULL,
                to="competition.Place",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="match",
            name="stage",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="matches",
                to="competition.Stage",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="match",
            name="stage_group",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="matches",
                verbose_name="Pool",
                blank=True,
                to="competition.StageGroup",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="laddersummary",
            name="stage",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="ladder_summary",
                to="competition.Stage",
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="laddersummary",
            name="stage_group",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="ladder_summary",
                to="competition.StageGroup",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="laddersummary",
            name="team",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="ladder_summary",
                to="competition.Team",
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name="laddersummary",
            unique_together=set([("stage", "team")]),
        ),
        migrations.AddField(
            model_name="ladderentry",
            name="match",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="ladder_entries",
                to="competition.Match",
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="ladderentry",
            name="opponent",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Team",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="ladderentry",
            name="team",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="ladder_entries",
                to="competition.Team",
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="ground",
            name="venue",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="grounds",
                to="competition.Venue",
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name="divisionexclusiondate",
            unique_together=set([("division", "date")]),
        ),
        migrations.AddField(
            model_name="division",
            name="season",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="divisions",
                to="competition.Season",
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="clubrole",
            name="competition",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="club_roles",
                to="competition.Competition",
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="clubassociation",
            name="person",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Person", on_delete=django.db.models.deletion.PROTECT
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="clubassociation",
            name="roles",
            field=touchtechnology.common.db.models.ManyToManyField(
                to="competition.ClubRole"
            ),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name="clubassociation",
            unique_together=set([("club", "person")]),
        ),
        migrations.AddField(
            model_name="club",
            name="primary",
            field=touchtechnology.common.db.models.ForeignKey(
                related_name="+club",
                blank=True,
                to="competition.Person",
                help_text="Appears on the front-end with other club information.",
                null=True,
                verbose_name="Primary contact",
                on_delete=django.db.models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
        migrations.CreateModel(
            name="ByeTeam",
            fields=[],
            options={
                "proxy": True,
            },
            bases=("competition.team",),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0022_competition__remove__person_id"),
    ]

    operations = [
        migrations.AddField(
            model_name="match",
            name="uuid",
            field=models.UUIDField(
                null=True,
                editable=False,
                unique=True,
                db_index=True,
            ),
        ),
    ]



from django.utils.functional import cached_property


class AppRouter(object):
    """
    Allow application level routing to distinct databases. Be careful not to
    attempt foreign-key relationships across physical boundaries!

    In your DJANGO_CONFIG_MODULE you should set an iterable mapping of
    app_label to database aliases. For example:

        TOUCHTECHNOLOGY_APP_ROUTING = (
            ('foo', 'bar'),
            ('baz', 'foo'),
        )

    or

        TOUCHTECHNOLOGY_APP_ROUTING = {
            'foo': 'bar',
            'baz': 'foo',
        }

    Any app_label which is ommitted will be routed to the DEFAULT_DB_ALIAS.
    """

    @cached_property
    def route(self):
        from collections import defaultdict

        from django.db import DEFAULT_DB_ALIAS

        from .default_settings import APP_ROUTING

        return defaultdict(lambda: DEFAULT_DB_ALIAS, APP_ROUTING)

    def _route(self, model, **hints):
        return self.route[model._meta.app_label]

    db_for_read = _route
    db_for_write = _route

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        return db == self.route[app_label]



# -*- coding: utf-8 -*-

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0023_competition__add__match_uuid_finalize"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="competition",
            options={"ordering": ("order",)},
        ),
        migrations.AlterModelOptions(
            name="division",
            options={"ordering": ("order",)},
        ),
        migrations.AlterModelOptions(
            name="season",
            options={"ordering": ("order",)},
        ),
    ]



import logging
import os.path
from importlib import import_module

from django.conf import settings
from django.core.cache import cache
from django.db.models.signals import post_save
from django.db.utils import DatabaseError
from django.utils.module_loading import import_string

from touchtechnology.common.models import SitemapNode
from touchtechnology.content.app_settings import TENANT_MEDIA_PUBLIC
from touchtechnology.content.models import Placeholder

logger = logging.getLogger(__name__)


def install_placeholder(app):
    application = import_module(app)
    for cls_name in getattr(application, "INSTALL", ()):
        path = "%s.sites.%s" % (app, cls_name)
        logger.debug('Installing application "{0}"'.format(path))
        site = import_string(path)()
        namespace = site.app_name
        try:
            placeholder, created = Placeholder.objects.get_or_create(
                path=path, namespace=namespace
            )
        except DatabaseError:
            logger.exception("Unable to install placeholder: %s", path)
        else:
            if created:
                logger.info(
                    'Registered new application "%s" (pk=%d)',
                    placeholder.path,
                    placeholder.pk,
                )
            else:
                logger.debug(
                    'Application "%s" (pk=%d) already ' "exists, skipping.",
                    placeholder.path,
                    placeholder.pk,
                )


def template_path(base, filename, *args):
    args = [arg for arg in args if arg]
    for index in range(len(args), 0, -1):
        yield os.path.join(base, os.path.join(*args[:index]), filename)
    yield os.path.join(base, filename)


def get_media_storage(request):
    if hasattr(request, "tenant"):
        get_public_schema_name = import_string(
            "tenant_schemas.utils.get_public_schema_name"
        )

        public = request.tenant.schema_name == get_public_schema_name()
        if not public or (public and not TENANT_MEDIA_PUBLIC):
            return os.path.join(settings.MEDIA_ROOT, request.tenant.domain_url)

    return settings.MEDIA_ROOT


def invalidate_sitemapnode_urlpatterns(sender, instance, **kwargs):
    """
    When a SitemapNode is changed, we need to invalidate the cache so that
    subsequent requests have a newly generated urlconf derived.

    We could add extra conditional checking - for example, if the slug or the
    keyword arguments to a Placeholder node have changed. This will be _far_
    less frequent than it would seem worthwhile to worry about.
    """
    logger.debug("SitemapNodeMiddleware cache requires invalidation.")
    middleware = import_module("touchtechnology.content.middleware")
    cache.delete(middleware.DEHYDRATED_URLPATTERNS_KEY)


post_save.connect(invalidate_sitemapnode_urlpatterns, SitemapNode)



import unittest
from datetime import datetime
from pathlib import Path
from zoneinfo import ZoneInfo

from django import forms
from django.test.utils import modify_settings, override_settings
from django.utils import timezone
from django.utils.encoding import smart_str
from test_plus import TestCase

from touchtechnology.common.forms.fields import SelectDateTimeField


class TestForm1(forms.Form):
    timestamp = SelectDateTimeField()


class TestForm2(forms.Form):
    timestamp = SelectDateTimeField(required=False)


class TimeZoneTests(TestCase):
    def get_html(self, filename):
        with open(Path(__file__).parent / "test_html" / filename) as fp:
            html = fp.read()
        return smart_str(html)

    def test_select_date_time_field_empty_input(self):
        """Test empty input handling"""
        form = TestForm1(data={})
        self.assertFormError(form, "timestamp", ["This field is required."])

    def test_select_date_time_field_empty_input_not_required(self):
        """Test empty input handling when field not required"""
        form = TestForm2(data={})
        self.assertFormError(form, "timestamp", [])

    def test_select_date_time_field_partial_date(self):
        """Test validation with only date input"""
        data = {"timestamp_0": "2023-07-04"}
        form = TestForm1(data=data)
        # FIXME: should error with ['Enter a valid time.', 'Enter a valid time zone.']
        self.assertFormError(form, "timestamp", ["This field is required."])

    def test_select_date_time_field_invalid_date(self):
        """Test validation with invalid date values"""
        data = {
            "timestamp_0": "2013-07-32",  # invalid day
            "timestamp_1": "7:49",
            "timestamp_2": "Australia/Sydney",
        }
        form = TestForm1(data=data)
        self.assertFormError(form, "timestamp", ["Enter a valid date."])

        data = {
            "timestamp_0": "2013-13-04",  # invalid month
            "timestamp_1": "7:49",
            "timestamp_2": "Australia/Sydney",
        }
        form = TestForm1(data=data)
        self.assertFormError(form, "timestamp", ["Enter a valid date."])

    def test_select_date_time_field_invalid_time(self):
        """Test validation with invalid time values"""
        data = {
            "timestamp_0": "2013-07-04",
            "timestamp_1": "24:49",  # invalid hour
            "timestamp_2": "Australia/Sydney",
        }
        form = TestForm1(data=data)
        self.assertFormError(form, "timestamp", ["Enter a valid time."])

        data = {
            "timestamp_0": "2013-07-04",
            "timestamp_1": "7:60",  # invalid minute
            "timestamp_2": "Australia/Sydney",
        }
        form = TestForm1(data=data)
        self.assertFormError(form, "timestamp", ["Enter a valid time."])

    def test_select_date_time_field_invalid_timezone(self):
        """Test validation with invalid timezone"""
        data = {
            "timestamp_0": "2013-07-04",
            "timestamp_1": "7:49",
            "timestamp_2": "Invalid/Timezone",  # invalid timezone
        }
        form = TestForm1(data=data)
        self.assertFormError(
            form,
            "timestamp",
            [
                "Select a valid choice. Invalid/Timezone is not one of the available choices."
            ],
        )

    def test_select_date_time_field_valid_input(self):
        """Test validation with valid input"""
        data = {
            "timestamp_0": "2013-07-04",
            "timestamp_1": "7:49",
            "timestamp_2": "Australia/Sydney",
        }
        form = TestForm1(data=data)
        self.assertFormError(form, "timestamp", [])

        expected = timezone.make_aware(
            datetime(2013, 7, 4, 7, 49),
            ZoneInfo("Australia/Sydney"),
        )
        self.assertEqual(form.cleaned_data["timestamp"], expected)

    def test_select_date_time_field_leap_year(self):
        """Test validation with leap year date"""
        data = {
            "timestamp_0": "2020-02-29",  # leap year
            "timestamp_1": "7:49",
            "timestamp_2": "Australia/Sydney",
        }
        form = TestForm1(data=data)
        self.assertFormError(form, "timestamp", [])

        expected = timezone.make_aware(
            datetime(2020, 2, 29, 7, 49),
            ZoneInfo("Australia/Sydney"),
        )
        self.assertEqual(form.cleaned_data["timestamp"], expected)

    def test_select_date_time_field_non_leap_year(self):
        """Test validation with non-leap year date"""
        data = {
            "timestamp_0": "2021-02-29",  # non-leap year
            "timestamp_1": "7:49",
            "timestamp_2": "Australia/Sydney",
        }
        form = TestForm1(data=data)
        self.assertFormError(form, "timestamp", ["Enter a valid date."])

    def test_select_date_time_field_missing_timezone(self):
        """Test validation with missing timezone"""
        data = {
            "timestamp_0": "2013-07-04",
            "timestamp_1": "7:49",
            # missing timestamp_2 (timezone)
        }
        form = TestForm1(data=data)
        # FIXME: would want this to return "Enter a valid time zone."
        self.assertFormError(form, "timestamp", ["This field is required."])

    @override_settings(TIME_ZONE="Australia/Sydney")
    def test_select_date_time_field_initial_tz(self):
        timestamp = timezone.make_aware(
            datetime(2013, 3, 24, 14, 30),
            ZoneInfo("Australia/Sydney"),
        )

        form = TestForm1(initial={"timestamp": timestamp})
        haystack = smart_str(form)
        for needle in [
            '<input type="text" name="timestamp_0" value="2013-03-24" required id="id_timestamp_0">',
            '<input type="text" name="timestamp_1" value="14:30:00" required id="id_timestamp_1">',
            '<option value="Australia/Sydney" selected>GMT+10:00 Australia/Sydney</option>',
        ]:
            with self.subTest(needle=needle):
                self.assertInHTML(needle, haystack)


@unittest.skip("FIXME: we don't use this widely but it is not passing")
@override_settings(ROOT_URLCONF="example_app.urls", TIME_ZONE="Europe/Paris")
@modify_settings(
    MIDDLEWARE={"append": "touchtechnology.common.middleware.timezone_middleware"}
)
class TestTimezoneMiddleware(TestCase):
    def test_set_timezone(self):
        "Ensure time zone is correct before, during and after a set request"
        self.assertEqual(
            timezone.get_current_timezone_name(),
            timezone.get_default_timezone_name(),
        )

        data = {"tz": "Australia/Sydney"}
        self.post("set-timezone", data=data, follow=True)

        self.assertEqual(timezone.get_current_timezone_name(), "Australia/Sydney")



# -*- coding: utf-8 -*-

from django.db import migrations, models

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Article",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("headline", models.CharField(max_length=150, verbose_name="Headline")),
                ("abstract", models.TextField(verbose_name="Abstract")),
                (
                    "published",
                    touchtechnology.common.db.models.DateTimeField(
                        help_text="Set a date & time in the future to schedule an announcement.",
                        verbose_name="Published",
                    ),
                ),
                ("slug", models.SlugField(max_length=150, verbose_name="Slug")),
                (
                    "slug_locked",
                    models.BooleanField(default=False, verbose_name="Locked Slug"),
                ),
                (
                    "byline",
                    models.CharField(max_length=75, verbose_name="Byline", blank=True),
                ),
                (
                    "keywords",
                    models.CharField(
                        max_length=255, verbose_name="Keywords", blank=True
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(default=True, verbose_name="Enabled"),
                ),
                (
                    "image",
                    models.ImageField(
                        upload_to="news", null=True, verbose_name="Image", blank=True
                    ),
                ),
            ],
            options={
                "ordering": ("-published",),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="ArticleContent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "label",
                    models.SlugField(
                        max_length=20,
                        verbose_name="CSS class",
                        choices=[(b"copy", "Copy")],
                    ),
                ),
                ("sequence", models.PositiveIntegerField(verbose_name="Sequence")),
                (
                    "copy",
                    touchtechnology.common.db.models.HTMLField(
                        verbose_name="Copy", blank=True
                    ),
                ),
                (
                    "article",
                    models.ForeignKey(
                        related_name="content",
                        verbose_name="Article",
                        to="news.Article",
                        on_delete=models.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": ("sequence",),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Category",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=75, verbose_name="Title")),
                (
                    "short_title",
                    models.CharField(
                        help_text="Used in navigation, a shorter alternative to the main title.",
                        max_length=50,
                        verbose_name="Short Title",
                        blank=True,
                    ),
                ),
                ("slug", models.SlugField(max_length=75, verbose_name="Slug")),
                (
                    "slug_locked",
                    models.BooleanField(default=False, verbose_name="Locked Slug"),
                ),
                (
                    "is_active",
                    models.BooleanField(default=True, verbose_name="Enabled"),
                ),
                (
                    "hidden_from_navigation",
                    models.BooleanField(
                        default=False,
                        help_text="When set to 'Yes' this object will still be available, but will not appear in menus.",
                        verbose_name="Hide from menus",
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name="article",
            name="categories",
            field=models.ManyToManyField(
                to="news.Category", verbose_name="Categories", blank=True
            ),
            preserve_default=True,
        ),
    ]



from rest_framework import serializers

from tournamentcontrol.competition import models
from tournamentcontrol.competition.rest.v1.viewsets import SlugViewSet


class ClubSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.Club
        fields = (
            "title",
            "short_title",
            "slug",
            "abbreviation",
            "url",
            "facebook",
            "twitter",
            "youtube",
            "website",
        )
        read_only_fields = ("slug", "abbreviation")
        extra_kwargs = {"url": {"lookup_field": "slug", "view_name": "v1:club-detail"}}


class ClubViewSet(SlugViewSet):
    queryset = models.Club.objects.all()
    serializer_class = ClubSerializer



import base64

import requests
from celery import shared_task
from googleapiclient.http import MediaInMemoryUpload

from tournamentcontrol.competition.models import Match, Stage
from tournamentcontrol.competition.utils import (
    generate_fixture_grid,
    generate_scorecards,
)


@shared_task
def generate_pdf_scorecards(
    match_pks, templates, extra_context, stage_pk=None, **kwargs
):
    matches = Match.objects.filter(pk__in=match_pks)
    stage = None
    if stage_pk is not None:
        stage = Stage.objects.get(pk=stage_pk)
    data = generate_scorecards(
        matches, templates, "pdf", extra_context, stage, **kwargs
    )
    # We can't JSON encode bytes, so we need to base64 encode the
    # PDF document before handing it back to the result backend.
    return base64.b64encode(data).decode("utf8")


@shared_task
def generate_pdf_grid(season, extra_context, date=None):
    dates = [date] if date is not None else None
    data: bytes = generate_fixture_grid(
        season,
        dates=dates,
        format="pdf",
        extra_context=extra_context,
        http_response=False,  # Get bytes back, not a response object
    )
    # We can't JSON encode bytes, so we need to base64 encode the
    # PDF document before handing it back to the result backend.
    return base64.b64encode(data).decode("utf8")


@shared_task
def set_youtube_thumbnail(match_pk):
    """
    Asynchronously use the Google YouTube Data API to set the thumbnail for
    the match specified.
    """
    obj = Match.objects.get(pk=match_pk)
    season = obj.stage.division.season

    img = requests.get(obj.live_stream_thumbnail or season.live_stream_thumbnail)
    img.raise_for_status()

    media_body = MediaInMemoryUpload(
        img.content,
        mimetype=img.headers["Content-Type"],
        resumable=True,
    )

    obj.stage.division.season.youtube.thumbnails().set(
        videoId=obj.external_identifier,
        media_body=media_body,
    ).execute()



from dateutil.rrule import DAILY, WEEKLY
from django.db import models
from django.utils.translation import gettext_lazy as _

GENDER_CHOICES = (
    ("M", _("Male")),
    ("F", _("Female")),
    ("X", _("Unspecified")),
)

SEASON_MODE_CHOICES = (
    (WEEKLY, _("Season")),
    (DAILY, _("Tournament")),
)

WIN_LOSE = {
    "W": _("Winner"),
    "L": _("Loser"),
}


class LiveStreamPrivacy(models.TextChoices):
    PUBLIC = "public", _("Public")
    PRIVATE = "private", _("Private")
    UNLISTED = "unlisted", _("Unlisted")



import textwrap

from django.test import TestCase

from tournamentcontrol.competition import utils


class UtilTests(TestCase):
    def test_round_robin_even(self):
        self.assertEqual(
            [[(1, 4), (2, 3)], [(1, 3), (4, 2)], [(1, 2), (3, 4)]],
            utils.round_robin(range(1, 5)),
        )

    def test_round_robin_odd(self):
        self.assertEqual(
            [[(1, 0), (2, 3)], [(1, 3), (0, 2)], [(1, 2), (3, 0)]],
            utils.round_robin(range(1, 4)),
        )

    def test_round_robin_format_even(self):
        self.assertEqual(
            textwrap.dedent(
                """
                ROUND
                1: 1 vs 4
                2: 2 vs 3
                ROUND
                3: 1 vs 3
                4: 4 vs 2
                ROUND
                5: 1 vs 2
                6: 3 vs 4
                """
            ).strip(),
            utils.round_robin_format(range(1, 5)),
        )

    def test_sum_dict_integer(self):
        d1 = utils.SumDict({"a": 1, "b": 2})
        d2 = utils.SumDict({"b": 1, "c": 2})
        self.assertEqual(d1 + d2, utils.SumDict({"a": 1, "b": 3, "c": 2}))

    def test_revpow(self):
        self.assertEqual(utils.revpow(2**4, 2), 4)



from django.urls import include, path
from example_app.models import Relative, TestDateTimeField

from touchtechnology.common.sites import Application


class TestDateTimeFieldSite(Application):
    def __init__(self, name="datetime", app_name="datetime", *args, **kwargs):
        super(TestDateTimeFieldSite, self).__init__(
            name=name, app_name=app_name, *args, **kwargs
        )

    def get_urls(self):
        urlpatterns = [
            path("", self.index, name="index"),
        ]
        return urlpatterns

    def index(self, request, *args, **kwargs):
        return self.generic_edit_related(request, TestDateTimeField, Relative)


class TestQueryStringSite(Application):
    def __init__(self, name="querystring", app_name="querystring", *args, **kwargs):
        super(TestQueryStringSite, self).__init__(
            name=name, app_name=app_name, *args, **kwargs
        )

    def get_urls(self):
        urlpatterns = [
            path("", self.index, name="index"),
        ]
        return urlpatterns

    def index(self, request, *args, **kwargs):
        queryset = TestDateTimeField.objects.order_by("pk")

        # naive filtering without a form - nasty, i know...
        year = request.GET.get("year")
        if year:
            queryset = queryset.filter(datetime__year=year)
            month = request.GET.get("month")
            if month:
                queryset = queryset.filter(datetime__month=month)

        return self.generic_list(request, TestDateTimeField, queryset=queryset)


class TestContextProcessorsSite(Application):
    def __init__(self, name="context", app_name="context", *args, **kwargs):
        super(TestContextProcessorsSite, self).__init__(
            name=name, app_name=app_name, *args, **kwargs
        )

    def get_urls(self):
        urlpatterns = [
            path("env", self.env, name="env"),
            path("site", self.site_view, name="site"),
            path("tz", self.tz, name="tz"),
        ]
        return urlpatterns

    def env(self, request, *args, **kwargs):
        return self.render(request, self.template_path("env.html"), {})

    def site_view(self, request, *args, **kwargs):
        return self.render(request, self.template_path("site.html"), {})

    def tz(self, request, *args, **kwargs):
        return self.render(request, self.template_path("tz.html"), {})


class TestPaginationSite(Application):
    def __init__(self, name="pagination", app_name="pagination", *args, **kwargs):
        super(TestPaginationSite, self).__init__(
            name=name, app_name=app_name, *args, **kwargs
        )

    def get_urls(self):
        urlpatterns = [
            path("setting-only/", self.setting_only, name="setting_only"),
            path("parameter/<int:paginate_by>/", self.parameter, name="parameter"),
        ]
        return urlpatterns

    def setting_only(self, request, *args, **kwargs):
        return self.generic_list(request, TestDateTimeField)

    def parameter(self, request, paginate_by, *args, **kwargs):
        return self.generic_list(request, TestDateTimeField, paginate_by=paginate_by)


class TestGenericViewsSite(Application):
    def __init__(self, name="generic", app_name="generic", *args, **kwargs):
        super(TestGenericViewsSite, self).__init__(
            name=name, app_name=app_name, *args, **kwargs
        )

    def get_urls(self):
        vanilla_urls = (
            [
                path("", self.list, name="list"),
                path("add/", self.edit, name="add"),
                path("<int:pk>/", self.detail, name="detail"),
                path("<int:pk>/edit/", self.edit, name="edit"),
                path("<int:pk>/delete/", self.delete, name="delete"),
                path("<int:pk>/perms/", self.perms, name="perms"),
                path("edit/", self.edit_multiple, name="edit"),
            ],
            self.app_name,
        )

        permissions_urls = (
            [
                path("", self.list_with_perms, name="list"),
                path("add/", self.edit_with_perms, name="add"),
                path("<int:pk>/", self.detail_with_perms, name="detail"),
                path("<int:pk>/edit/", self.edit_with_perms, name="edit"),
                path("<int:pk>/delete/", self.delete_with_perms, name="delete"),
                path("edit/", self.edit_multiple_with_perms, name="edit"),
            ],
            self.app_name,
        )

        urlpatterns = [
            path("vanilla/", include(vanilla_urls, namespace="vanilla")),
            path("permissions/", include(permissions_urls, namespace="permissions")),
        ]
        return urlpatterns

    def list(self, request, **extra_context):
        return self.generic_list(
            request, TestDateTimeField, extra_context=extra_context
        )

    def detail(self, request, pk, **extra_context):
        return self.generic_detail(
            request, TestDateTimeField, pk=pk, extra_context=extra_context
        )

    def edit(self, request, pk=None, **extra_context):
        return self.generic_edit(
            request, TestDateTimeField, pk=pk, extra_context=extra_context
        )

    def delete(self, request, pk, **kwargs):
        return self.generic_delete(request, TestDateTimeField, pk=pk, **kwargs)

    def perms(self, request, pk, **extra_context):
        return self.generic_permissions(
            request, TestDateTimeField, pk=pk, extra_context=extra_context
        )

    def edit_multiple(self, request, **extra_context):
        return self.generic_edit_multiple(
            request, TestDateTimeField, extra_context=extra_context
        )

    def list_with_perms(self, request, **extra_context):
        return self.generic_list(
            request,
            TestDateTimeField,
            permission_required=True,
            extra_context=extra_context,
        )

    def edit_with_perms(self, request, pk=None, **extra_context):
        return self.generic_edit(
            request,
            TestDateTimeField,
            pk=pk,
            permission_required=True,
            extra_context=extra_context,
        )

    def delete_with_perms(self, request, pk, **kwargs):
        return self.generic_delete(
            request, TestDateTimeField, pk=pk, permission_required=True, **kwargs
        )

    def edit_multiple_with_perms(self, request, **extra_context):
        return self.generic_edit_multiple(
            request,
            TestDateTimeField,
            permission_required=True,
            extra_context=extra_context,
        )

    def detail_with_perms(self, request, pk, **extra_context):
        return self.generic_detail(
            request,
            TestDateTimeField,
            pk=pk,
            # permission_required=True,
            extra_context=extra_context,
        )



from test_plus import TestCase

from touchtechnology.common.tests import factories


class SitemapMoveTest(TestCase):

    def test_move_down_up(self):
        user = factories.UserFactory.create(is_staff=True, is_superuser=True)
        root = factories.SitemapNodeFactory.create()
        factories.SitemapNodeFactory.create_batch(parent=root, size=4)

        # baseline for the mptt values of the nodes, order by pk to ensure
        # subsequent check is valid
        before_move = sorted(
            root.get_children().values_list("pk", "lft", "rght", "tree_id")
        )

        # perform the move using the admin view
        with self.login(user):
            self.get("admin:reorder", before_move[0][0], "down")
            self.response_302()

            # after moving the node down make sure the mptt values are different
            after_move_1 = sorted(
                root.get_children().values_list("pk", "lft", "rght", "tree_id")
            )
            self.assertNotEqual(before_move, after_move_1)

            # try moving the node back up
            self.get("admin:reorder", before_move[0][0], "up")
            self.response_302()

            # after moving the node up the mptt values should be restored
            after_move_2 = sorted(
                root.get_children().values_list("pk", "lft", "rght", "tree_id")
            )
            # self.assertNotEqual(after_move_1, after_move_2)
            self.assertEqual(before_move, after_move_2)



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0036_auto_20191122_1340"),
    ]

    operations = [
        migrations.AlterField(
            model_name="match",
            name="evaluated",
            field=models.BooleanField(null=True),
        ),
    ]



import logging
from calendar import timegm

from django.conf import settings
from django.db.models import Max
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.utils.http import http_date, parse_http_date_safe
from django.views.decorators.cache import patch_cache_control
from django.views.decorators.http import last_modified

from touchtechnology.common.models import SitemapNode
from touchtechnology.content.models import Page
from touchtechnology.content.utils import template_path

logger = logging.getLogger(__name__)


def page_last_modified(request, page_id=None, **kwargs):
    last_modified = SitemapNode.objects.aggregate(last_modified=Max("last_modified"))[
        "last_modified"
    ]
    logger.debug(
        'SitemapNode.last_modified.max="%s"',
        http_date(timegm(last_modified.utctimetuple())),
    )
    return last_modified


@last_modified(page_last_modified)
def dispatch(request, page_id=None, node=None, url=None):
    if page_id is not None:
        page = get_object_or_404(Page, pk=page_id)
    else:
        page = None

    if page and settings.DEBUG:
        if_modified_since = request.META.get("HTTP_IF_MODIFIED_SINCE")
        if if_modified_since:
            if_modified_since = parse_http_date_safe(if_modified_since)
            res_last_modified = timegm(page.last_modified.utctimetuple())
            logger.debug(
                'page="%d", if-last-modified="%s", ' 'last-modified="%s"',
                page_id,
                http_date(if_modified_since),
                http_date(res_last_modified),
            )

    base = "touchtechnology/content"
    args = url.split("/")
    templates = list(
        template_path(base, "%s.html" % (page and "page" or "folder"), *args)
    )

    if page and page.template:
        templates.insert(0, page.template)

    context = {
        "page": page,
        "node": node,
    }
    response = TemplateResponse(request, templates, context)

    # For authenticated visitors we may alter the rendered content, so mark
    # the content as private to stop interim caches from keeping a copy.
    if request.user.is_authenticated:
        patch_cache_control(response, private=True)
    return response



# -*- coding: utf-8 -*-

from django.db import migrations, models

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("common", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Chunk",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("slug", models.SlugField(verbose_name="Slug")),
                (
                    "copy",
                    touchtechnology.common.db.models.HTMLField(
                        verbose_name="Page Copy", blank=True
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Content",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("copy", touchtechnology.common.db.models.HTMLField(blank=True)),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="NodeContent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "copy",
                    touchtechnology.common.db.models.HTMLField(
                        verbose_name="Page Copy", blank=True
                    ),
                ),
                (
                    "node",
                    models.ForeignKey(
                        related_name="contents",
                        verbose_name="Node",
                        to="common.SitemapNode",
                        on_delete=models.PROTECT,
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Page",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "template",
                    touchtechnology.common.db.models.TemplatePathField(
                        max_length=200, verbose_name="Template", blank=True
                    ),
                ),
                (
                    "keywords",
                    models.CharField(
                        help_text="This should be a comma-separated list of terms that indicate the content of this page - used to assist Search Engines rank your page.",
                        max_length=255,
                        verbose_name="Keywords",
                        blank=True,
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        help_text="Search Engines may use this when determining the relevance of your page.",
                        max_length=255,
                        verbose_name="Description",
                        blank=True,
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="PageContent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "label",
                    models.SlugField(
                        default="copy",
                        max_length=20,
                        verbose_name="CSS class",
                        choices=[(b"copy", b"copy")],
                    ),
                ),
                ("sequence", models.PositiveIntegerField(verbose_name="Sequence")),
                (
                    "copy",
                    touchtechnology.common.db.models.HTMLField(
                        verbose_name="Copy", blank=True
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        related_name="content",
                        verbose_name="Page",
                        to="content.Page",
                        on_delete=models.PROTECT,
                    ),
                ),
            ],
            options={
                "ordering": ("sequence",),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Placeholder",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("path", models.CharField(max_length=255, verbose_name="Module path")),
                (
                    "namespace",
                    models.CharField(
                        max_length=255, verbose_name="Namespace", db_index=True
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="PlaceholderKeywordArgument",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("key", models.CharField(max_length=200, verbose_name="Key")),
                (
                    "value",
                    models.TextField(
                        default="null",
                        help_text="Enter a valid JSON object",
                        verbose_name="Value",
                        blank=True,
                    ),
                ),
                (
                    "node",
                    models.ForeignKey(
                        related_name="kw",
                        verbose_name="Node",
                        to="common.SitemapNode",
                        on_delete=models.PROTECT,
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Redirect",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "source_url",
                    models.CharField(
                        help_text="The path that will trigger the redirection.",
                        max_length=250,
                        verbose_name="Source URL",
                    ),
                ),
                (
                    "destination_url",
                    models.CharField(
                        help_text="The URL or path that the browser will be sent to.",
                        max_length=500,
                        verbose_name="Destination URL",
                    ),
                ),
                (
                    "label",
                    models.CharField(max_length=100, verbose_name="Label", blank=True),
                ),
                ("active", models.BooleanField(default=True, verbose_name="Active")),
                (
                    "permanent",
                    models.BooleanField(default=False, verbose_name="Permanent"),
                ),
            ],
            options={
                "ordering": ("destination_url",),
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name="placeholderkeywordargument",
            unique_together=set([("node", "key")]),
        ),
    ]



import functools
import logging
from operator import attrgetter, or_
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from django.apps import apps
from django.core.cache import cache
from django.core.paginator import Paginator
from django.db import connection
from django.db.models import Model, Q
from django.http import Http404, HttpRequest
from django.template.loader import select_template
from first import first

from touchtechnology.common.exceptions import NotModelManager
from touchtechnology.common.mixins import NodeRelationMixin

logger = logging.getLogger(__name__)


class FauxNodeMeta(object):
    left_attr = "lft"
    right_attr = "rght"
    level_attr = "level"


class FauxNode(NodeRelationMixin):
    REPR_FMT = "<{class}: {title} ({tree_id}:{lft},{rght})>"

    def __init__(
        self,
        app,
        title,
        short_title,
        slug,
        level,
        tree_id,
        lft=None,
        rght=None,
        parent=None,
        changefreq="monthly",
        priority=0.5,
    ):
        self.app = app
        self.title = title
        self.short_title = short_title
        self.slug = slug
        self.level = level
        self.tree_id = tree_id
        self.lft = lft
        self.rght = rght
        if parent is None and app is not None:
            self.parent = app.node
        else:
            self.parent = parent

        try:
            self.changefreq = changefreq
        except AttributeError:
            logger.exception("Unable to set `changefreq` for %r", self)

        try:
            self.priority = priority
        except AttributeError:
            logger.exception("Unable to set `priority` for %r", self)

        self._mptt_meta = FauxNodeMeta()

    def __eq__(self, other):
        if other is None:
            return
        attr = ("title", "short_title", "slug", "level", "tree_id", "parent")
        gttr = [attrgetter(a) for a in attr]
        return all([g(self) == g(other) for g in gttr])

    def __repr__(self):
        attr = (
            "title",
            "short_title",
            "slug",
            "level",
            "tree_id",
            "lft",
            "rght",
            "parent",
        )
        gttr = [attrgetter(a) for a in attr]
        data = dict(zip(attr, [g(self) for g in gttr]))
        data["class"] = self.__class__.__name__
        return self.REPR_FMT.format(**data)

    @classmethod
    def equal(cls, iterable, find):
        return first(iterable, key=lambda i: i == find)

    def get_absolute_url(self):
        raise NotImplementedError

    def get_ancestors(self):
        n = self
        while n.parent:
            yield n.parent
            n = n.parent

    def get_root(self):
        if self.parent is None:
            return self
        return self.parent.get_root()


def get_mod_func(callback):
    # Converts 'touchtechnology.common.utils.get_mod_func' to
    # ['touchtechnology.common.utils', 'get_mod_func']
    try:
        dot = callback.rindex(".")
    except ValueError:
        return callback, ""
    return callback[:dot], callback[dot + 1 :]


def create_exclude_filter(queryset):
    def _filter(node):
        return Q(tree_id=node.tree_id, lft__gte=node.lft, lft__lte=node.rght)

    return functools.reduce(or_, [_filter(n) for n in queryset], Q())


def get_perms_for_model(model, add=False, change=False, delete=False, *extra):
    tpl = "%s.%%s_%s" % (model._meta.app_label, model._meta.model_name)
    res = []
    if add:
        res.append(tpl % "add")
    if change:
        res.append(tpl % "change")
    if delete:
        res.append(tpl % "delete")
    for x in extra:
        res.append(tpl % x)
    return res


def get_all_perms_for_model(model, *extra):
    return get_perms_for_model(model, True, True, True, *extra)


def tree_for_node(node):
    """
    Utility function to construct a QuerySet which has the expanded set
    of nodes for the given node.
    """
    from .models import SitemapNode

    if not isinstance(node, SitemapNode):
        real_nodes = SitemapNode.objects.filter(tree_id=node.tree_id, lft__lte=node.lft)
        node = first(real_nodes.reverse())

    if node:
        parents = node.get_ancestors()
        siblings = node.get_siblings(include_self=True)
        children = node.get_descendants()

        rest = functools.reduce(
            or_, [p.get_siblings() for p in parents], SitemapNode.objects.none()
        )

        nodes = parents | siblings | children | rest
        nodes = nodes.filter(level__lte=node.level + 1)

    else:
        nodes = SitemapNode.objects.none()

    return nodes


def model_and_manager(model_or_manager):
    """
    All our generic views work on the basis of a ``model_or_manager`` which
    needs to be unpacked into a ``model`` and ``manager``.

    The function will reduce boilerplate in generic methods.
    """
    if isinstance(model_or_manager, Model):
        model = model_or_manager._meta.model
        manager = model._default_manager
    elif hasattr(model_or_manager, "model"):
        model = model_or_manager.model
        manager = model_or_manager
    else:
        try:
            model = model_or_manager
            manager = model_or_manager._default_manager
        except AttributeError:
            raise NotModelManager(model_or_manager)
    return model, manager


def determine_page_number(request, paginator):
    """
    Determine what page number for the current request.

        request:    django.http.HttpRequest instance
        paginator:  django.core.paginator.Paginator instance
    """
    assert isinstance(request, HttpRequest)
    assert isinstance(paginator, Paginator)
    page = request.GET.get("page", 1)
    try:
        page_number = int(page)
    except ValueError:
        if page == "last":
            page_number = paginator.num_pages
        else:
            raise Http404
    return page_number


def select_template_name(templates):
    """
    Return the template name that wins template resolution for an iterable of
    template names.
    """
    template = select_template(templates)
    return template.template.name


def get_timezone_from_request(request):
    if hasattr(request, "session"):
        tzname = request.session.get("django_timezone", None)
        if tzname is not None:
            try:
                tzinfo = ZoneInfo(tzname)
            except ZoneInfoNotFoundError:
                tzinfo = None
            return tzinfo


def get_base_url(scheme="http"):
    """
    It is frustratingly difficult to obtain the appropriate base URL to use
    for sending links outside of the request-response cycle. If we are using
    the django-tenant-schemas application, this can be somewhat simplified.
    If we are not, fall back to using the django.contrib.sites framework.
    """
    if hasattr(connection, "tenant"):
        hostname = connection.tenant.domain_url
        # We have our own pattern of using the `prepend_www` attribute so lets
        # check for it and stick it on the front.
        if getattr(connection.tenant, "prepend_www", False):
            hostname = "www." + hostname
    else:
        site = apps.get_model("sites.Site").objects.get_current()
        hostname = site.domain
    context = {
        "scheme": scheme,
        "hostname": hostname,
    }
    return "%(scheme)s://%(hostname)s/" % context


def get_all_perms_for_model_cached(model, ttl=60, **extra):
    """
    Private function to cache and return the list of permissions that can be
    set on a given model.
    """
    # What are the permissions that this model accepts?
    cache_key = "model_perms.%s.%s" % (model._meta.app_label, model._meta.model_name)
    model_perms = cache.get(cache_key)
    logger.debug("model_perms_cache=%s", "MISS" if model_perms is None else "HIT")
    if model_perms is None:
        model_perms = get_all_perms_for_model(model, **extra)
        cache.set(cache_key, model_perms, timeout=ttl)
    return model_perms



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("competition", "0044_match_live_stream_bind"),
    ]

    operations = [
        migrations.AddField(
            model_name="ground",
            name="stream_key",
            field=models.CharField(
                blank=True, db_index=True, max_length=50, null=True, unique=True
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [("competition", "0033_matchscoresheet")]

    operations = [
        migrations.AddField(
            model_name="division",
            name="copy",
            field=touchtechnology.common.db.models.HTMLField(blank=True),
        ),
        migrations.AddField(
            model_name="place",
            name="copy",
            field=touchtechnology.common.db.models.HTMLField(blank=True),
        ),
        migrations.AddField(
            model_name="rankdivision",
            name="copy",
            field=touchtechnology.common.db.models.HTMLField(blank=True),
        ),
        migrations.AddField(
            model_name="season",
            name="copy",
            field=touchtechnology.common.db.models.HTMLField(blank=True),
        ),
        migrations.AddField(
            model_name="stage",
            name="copy",
            field=touchtechnology.common.db.models.HTMLField(blank=True),
        ),
        migrations.AddField(
            model_name="stagegroup",
            name="copy",
            field=touchtechnology.common.db.models.HTMLField(blank=True),
        ),
        migrations.AddField(
            model_name="team",
            name="copy",
            field=touchtechnology.common.db.models.HTMLField(blank=True),
        ),
    ]



import logging
from email.utils import formataddr

from django.conf import settings
from django.core.mail import send_mail
from django.template import Context, Template

from tournamentcontrol.competition.calc import BonusPointCalculator, Calculator
from tournamentcontrol.competition.signals.decorators import (
    disable_for_loaddata,
)
from tournamentcontrol.competition.utils import forfeit_notification_recipients

logger = logging.getLogger(__name__)


@disable_for_loaddata
def match_saved_handler(sender, instance, created, *args, **kwargs):
    """
    Function to be called following a Match being saved.

    This should create the necessary LadderEntry objects for the match in
    question (by deleting previous entries) and inserting new values.
    """
    for ladder_entry in instance.ladder_entries.all():
        ladder_entry.delete()

    if instance.is_bye and instance.bye_processed:
        logger.debug("BYE: Match #%s", instance.pk)
        return create_match_ladder_entries(instance)

    elif instance.is_forfeit:
        logger.debug("FORFEIT: Match #%s", instance.pk)
        return create_match_ladder_entries(instance)

    elif instance.home_team_score is not None and instance.away_team_score is not None:
        logger.debug("RESULT: Match #%s", instance.pk)
        return create_match_ladder_entries(instance)

    logger.debug("SKIPPED: Match #%s", instance.pk)


def match_deleted_handler(sender, instance, *args, **kwargs):
    """
    Function to be called prior to a Match being saved.

    This should remove the related LadderEntry objects for the match in
    question.
    """
    instance.ladder_entries.all().delete()


def create_match_ladder_entries(instance):
    home_ladder = create_team_ladder_entry(instance, "home")
    away_ladder = create_team_ladder_entry(instance, "away")
    return dict(home_ladder=home_ladder, away_ladder=away_ladder)


def create_team_ladder_entry(instance, home_or_away):
    from tournamentcontrol.competition.models import LadderEntry

    if home_or_away == "home":
        opponent = "away"
    elif home_or_away == "away":
        opponent = "home"

    team = getattr(instance, home_or_away + "_team", None)
    team_score = getattr(instance, home_or_away + "_team_score") or 0
    opponent_obj = getattr(instance, opponent + "_team", None)
    opponent_score = getattr(instance, opponent + "_team_score") or 0

    ladder_kwargs = dict(
        match_id=instance.pk,
        played=1,
        score_for=team_score,
        score_against=opponent_score,
    )

    if team is not None:
        ladder_kwargs["team_id"] = team.pk

    if opponent_obj is not None:
        ladder_kwargs["opponent_id"] = opponent_obj.pk

    if instance.is_bye:
        ladder_kwargs["bye"] = 1

    elif instance.is_forfeit:
        if not instance.stage.division.include_forfeits_in_played:
            ladder_kwargs["played"] = 0
        ladder_kwargs["forfeit_for"] = int(team == instance.forfeit_winner)
        ladder_kwargs["forfeit_against"] = int(team != instance.forfeit_winner)

    else:
        ladder_kwargs["win"] = int(team_score > opponent_score)
        ladder_kwargs["loss"] = int(team_score < opponent_score)
        ladder_kwargs["draw"] = int(team_score == opponent_score)

    if team is not None:
        ladder = LadderEntry(**ladder_kwargs)

        calculator = Calculator(ladder)
        calculator.parse(instance.stage.division.points_formula)
        ladder.points = calculator.evaluate()

        if instance.stage.division.bonus_points_formula:
            bonus_calculator = BonusPointCalculator(ladder)
            bonus_calculator.parse(instance.stage.division.bonus_points_formula)
            ladder.bonus_points = bonus_calculator.evaluate()

        ladder.points += ladder.bonus_points
        ladder.save()

        return ladder


def notify_match_forfeit_email(sender, match, team, *args, **kwargs):
    """
    When a match is notified as having been forfeit, send a notification email
    to players in the opposition team and the designated season administrators.

    :param match: the match that was forfeit
    :param team: the team that forfeit the match
    :return: None
    """
    participants, administrators = forfeit_notification_recipients(match)

    context = Context(
        {
            "match": match,
            "team": team,
        }
    )

    # construct and send an email to the players in the opposition
    subject = Template(
        "Your {{ match.time }} game against {{ team.title }} " "has been forfeit"
    ).render(context)
    message = ""
    from_email = settings.DEFAULT_FROM_EMAIL
    recipient_list = [formataddr((p.get_full_name(), p.email)) for p in participants]

    send_mail(subject, message, from_email, recipient_list)

    # construct and send an email to the competition administrators
    subject = Template(
        "Forfeit: {{ match }} [{{ match.time }}, "
        "{{ match.date }}, {{ match.play_at }}]"
    ).render(context)
    recipient_list = [formataddr((p.get_full_name(), p.email)) for p in administrators]

    send_mail(subject, message, from_email, recipient_list)
